import matplotlib.pyplot as plt
import seaborn as sns
from .statistical_tests import *
import pandas as pd
import scipy.stats
from scipy.stats import norm
import math
import numpy as np
from scipy.optimize import minimize

import datetime
import os
import numpy as np
import pandas as pd
import empyrical as ep
import plotly.express as px
import plotly.graph_objects as go
import plotly.figure_factory as ff

from scipy.optimize import minimize
from plotly.subplots import make_subplots

import plotly.io as pio
pio.templates.default = "simple_white"


def plot_efficient_frontier(df:pd.DataFrame, frequency:str="D", num_ports:int=10000, risk_free_rate:int=0, colorscale:str="aggrnyl", ef_line:bool=False, custom_weights:list=[], graph_width:int=1600, graph_height:int=600) -> go.Figure:
    '''
    Plots the Markowitz Efficient Frontier given a dataframe that contains the asset prices
    for a given period.

    Parameters
    ----------
    df : Pandas DataFrame object
        DataFrame containing the asset prices
    frequency : str, default 'D'    Frequency of the data
    num_ports : int, default 10000
        The number of portfolios to generate
    risk_free_rate : str, default 0
        Market's risk free rate
    colorscale : Color map of the efficient frontier, default 'aggrnyl'
    ef_line: bool, default False
        Plots only the Efficient Frontier Line
    custom_weights: np.array or list, default []
        custom weights to be plotted on the Efficient Frontier
    strategy : str, default - blank string
        Name of the strategy
    graph_width : int, default - 1600
        Width of the graph object
    graph_height : int, default - 600
        Height of the graph object

    Returns
    -------
    plotly.graph_objects : Plots the Efficient Frontier

    Example
    -------
    >>> df.head(2)

                    TCS  NTPC  INFY
    Dates
    2022-01-01      200  300   400
    2022-01-02      202  299   401


    >>> plot_efficient_frontier(df)
    '''
    # Type checking
    assert (str(type(df)) == "<class 'pandas.core.frame.DataFrame'>"), "Not a dataframe"
    assert (frequency in ['D', 'M', 'Q', 'W', 'Y', '6M']), "Should be one of the values of 'D', 'M', 'Q', 'W', 'Y', '6M'"
    assert (num_ports > 1), "Number of portfolio cannot be less than 1"
    assert (risk_free_rate >= 0), "risk free rate cannot be negative"
    assert (str(type(ef_line))=="<class 'bool'>"), "Not a boolean variable"
    assert (str(type(graph_width)) == "<class 'int'>"), "Not an integer value"
    assert (str(type(graph_height)) == "<class 'int'>"), "Not an integer value"

    try:
        # Mapping the frequencies
        mapper = {'D': 252, 'M': 12, 'W': 52, 'Q': 4, 'Y': 1, '6M': 2}

        # Calculating the annualized returns and volatilities
        rets = df.pct_change().dropna().mean() * mapper[frequency]
        cov = df.pct_change().dropna().cov() * mapper[frequency]

        # Only efficient frontier portfolios
        mean_variance_ef_portfolios = []

        # Generating various portfolios
        np.random.seed(42)
        all_weights = np.zeros((num_ports, len(df.columns)))
        ret_arr = np.zeros(num_ports)
        vol_arr = np.zeros(num_ports)
        sharpe_arr = np.zeros(num_ports)

        # Lists to store various weight combinations and assets.
        all_weights_plot = []
        all_ticker_plot = []


        # Plot all portfolio / Plot only efficient portfolios
        if ef_line:

            for x in range(num_ports):
                # Weights
                weights = np.array(np.random.random(len(df.columns)))
                weights = weights / np.sum(weights)

                # Save weights
                all_weights[x, :] = weights

                # Expected return
                ret_arr[x] = np.sum((rets * weights))

                # Expected volatility
                vol_arr[x] = np.sqrt(np.dot(weights.T, np.dot(cov, weights)))

                # Sharpe Ratio
                sharpe_arr[x] = (ret_arr[x] - risk_free_rate) / vol_arr[x]

                track = False
                # Check for optimal portfolios
                for v, r, _ in mean_variance_ef_portfolios:
                    if (r > ret_arr[x]) and (v < vol_arr[x]):
                        track = True
                        break

                if track != True:
                    # Storing weights and tickers for portfolio plots
                    all_weights_plot.append(all_weights[x])
                    all_ticker_plot.append(list(df.columns))
                    mean_variance_ef_portfolios.append([vol_arr[x], ret_arr[x], sharpe_arr[x]])


            # Max seen return in simulation
            indices_max_r = max(np.array(mean_variance_ef_portfolios)[:, 1])

            # Min seen return in simulation
            indices_min_r = min(np.array(mean_variance_ef_portfolios)[:, 1])


            # Efficient Frontier Line Calculations
            def checkSumToOne(w: np.array):
                return np.sum(w) - 1

            def getReturn(w: np.array):
                w = np.array(w)
                R = np.sum(rets * w)
                return R

            def minimizeMyVolatility(w: np.array):
                w = np.array(w)
                V = np.sqrt(np.dot(w.T, np.dot(cov, w)))
                return V

            def get_efficient_frontier():
                returns = np.linspace((indices_min_r - 0.05), indices_max_r, 20)
                volatility_opt = []

                # Defining starting weights (equal weights)
                w0 = [1/len(rets)] * len(rets)

                # Defining bounds
                if len(rets) == 1:
                    bounds = (0, 1)
                else:
                    bounds = ((0, 1), ) * len(rets)


                for ret in returns:
                    # Find best volatility

                    # Constraints
                    constraints = ({'type':'eq', 'fun':checkSumToOne},
                                   {'type': 'eq', 'fun': lambda w: getReturn(w) - ret})

                    opt = minimize(minimizeMyVolatility, w0, method='SLSQP', bounds=bounds, constraints=constraints)

                    # Save optimal volatility
                    volatility_opt.append(opt['fun'])

                return volatility_opt, returns


            # EF Line Volatilities, Returns
            volas, retus = get_efficient_frontier()


            # Type casting
            mean_variance_ef_portfolios = np.array(mean_variance_ef_portfolios)

            # Max Sharpe Portfolio
            indices_max_sr = mean_variance_ef_portfolios[:, 2].argmax()
            max_sr_ret = mean_variance_ef_portfolios[:, 1][indices_max_sr]
            max_sr_vol = mean_variance_ef_portfolios[:, 0][indices_max_sr]
            max_sr_weights = all_weights_plot[indices_max_sr]


            # Min Vol Portfolio
            indices_min_vol = mean_variance_ef_portfolios[:, 0].argmin()
            min_vol_ret = mean_variance_ef_portfolios[:, 1][indices_min_vol]
            min_vol_vol = mean_variance_ef_portfolios[:, 0][indices_min_vol]
            min_vol_weights = all_weights_plot[indices_min_vol]

            # Plot the efficient frontier
            fig = go.Figure()

            fig.add_trace(go.Scatter(
                y = mean_variance_ef_portfolios[:, 1],
                x = mean_variance_ef_portfolios[:, 0],
                mode = 'markers',
                name = 'Portfolios',
                marker = dict(color=mean_variance_ef_portfolios[:, 2],
                showscale = True,
                size = 5,
                line = dict(width=1),
                colorscale = colorscale,
                colorbar = dict(title="Sharpe<br>Ratio")),
                text = [str(np.array(all_ticker_plot[i])) + "<br>" + str(np.array(all_weights_plot[i]).round(2)) for i in range(len(all_ticker_plot))]
            ))

            # Max Sharpe Plot
            fig.add_trace(go.Scatter(
                y = [max_sr_ret],
                x = [max_sr_vol],
                mode = 'markers',
                marker_symbol = 'star',
                marker = dict(color='red', size=16),
                text = str(np.array(all_ticker_plot[indices_max_sr])) + "<br>" + str(np.array(max_sr_weights).round(2)),
                name = "Maximum Sharpe Ratio Portfolio"
            ))

            # Min Vol Plot
            fig.add_trace(go.Scatter(
                y = [min_vol_ret],
                x = [min_vol_vol],
                mode = 'markers',
                marker_symbol = 'diamond',
                marker = dict(color='black', size=16),
                text = str(np.array(all_ticker_plot[indices_min_vol])) + "<br>" + str(np.array(min_vol_weights).round(2)),
                name = "Minimum Volatility Portfolio"
            ))

            # Only Line
            fig.add_trace(go.Scatter(
                y = retus,
                x = volas,
                mode = 'lines',
                line = dict(shape = 'linear', color = 'rgb(10, 120, 24)', dash = 'dot'),
                name = "Efficient Frontier Line"
            ))


            # custom portfolios
            ret_custom, vol_custom, stocks_custom, wts_custom = [], [], [], []

            if len(custom_weights) > 0:
                for wts in custom_weights:
                    wts = np.array(wts)
                    ret_custom.append(np.sum(wts * rets))
                    vol_custom.append(np.sqrt(np.dot(wts.T, np.dot(cov, wts))))
                    stocks_custom.append(list(df.columns))
                    wts_custom.append(wts)

            if len(custom_weights) > 0:
                fig.add_trace(go.Scatter(
                    y = ret_custom,
                    x = vol_custom,
                    mode = 'markers',
                    name = 'Custom Portfolios',
                    marker = dict(color='blue', size=16),
                    marker_symbol = 'triangle-up',
                    text = [str(np.array(stocks_custom[i])) + "<br>" + str(np.array(wts_custom[i]).round(2)) for i in range(len(stocks_custom))]
                ))

            # Add title/labels
            fig.update_layout(template='plotly_white',
                            xaxis=dict(title='Annualised Risk (Volatility)'),
                            yaxis=dict(title='Annualised Return'),
                            title='Markowitz Efficient Frontier',
                            coloraxis_colorbar=dict(title="Sharpe Ratio"),
                            legend=dict(yanchor="bottom", y=0.1, xanchor="right", x=0.99),
                            font=dict(family = "Courier New, monospace",
                                      size = 12,
                                      color="RebeccaPurple"
                                    ), width = graph_width, height = graph_height)

            return fig

        else:

            for x in range(num_ports):
                # Weights
                weights = np.array(np.random.random(len(df.columns)))
                weights = weights / np.sum(weights)

                # Save weights
                all_weights[x, :] = weights

                # Expected return
                ret_arr[x] = np.sum((rets * weights))

                # Expected volatility
                vol_arr[x] = np.sqrt(np.dot(weights.T, np.dot(cov, weights)))

                # Sharpe Ratio
                sharpe_arr[x] = (ret_arr[x] - risk_free_rate) / vol_arr[x]

                # Storing weights and tickers for portfolio plots
                all_weights_plot.append(all_weights[x])
                all_ticker_plot.append(list(df.columns))

            # Max Sharpe Portfolio
            indices_max_sr = sharpe_arr.argmax()
            max_sr_ret = ret_arr[indices_max_sr]
            max_sr_vol = vol_arr[indices_max_sr]
            max_sr_weights = all_weights[indices_max_sr]


            # Min Vol Portfolio
            indices_min_vol = vol_arr.argmin()
            min_vol_ret = ret_arr[indices_min_vol]
            min_vol_vol = vol_arr[indices_min_vol]
            min_vol_weights = all_weights[indices_min_vol]


            # Plot the efficient frontier
            fig = go.Figure()

            fig.add_trace(go.Scatter(
                y = ret_arr,
                x = vol_arr,
                mode = 'markers',
                name = 'Portfolios',
                marker = dict(color=sharpe_arr,
                showscale = True,
                size = 5,
                line = dict(width=1),
                colorscale = colorscale,
                colorbar = dict(title="Sharpe<br>Ratio")),
                text = [str(np.array(all_ticker_plot[i])) + "<br>" + str(np.array(all_weights_plot[i]).round(2)) for i in range(len(all_ticker_plot))]
            ))

            # Max Sharpe Plot
            fig.add_trace(go.Scatter(
                y = [max_sr_ret],
                x = [max_sr_vol],
                mode = 'markers',
                marker_symbol = 'star',
                marker = dict(color='red', size=16),
                text = str(np.array(all_ticker_plot[indices_max_sr])) + "<br>" + str(np.array(max_sr_weights).round(2)),
                name = "Maximum Sharpe Ratio Portfolio"
            ))

            # Min Vol Plot
            fig.add_trace(go.Scatter(
                y = [min_vol_ret],
                x = [min_vol_vol],
                mode = 'markers',
                marker_symbol = 'diamond',
                marker = dict(color='black', size=16),
                text = str(np.array(all_ticker_plot[indices_min_vol])) + "<br>" + str(np.array(min_vol_weights).round(2)),
                name = "Minimum Volatility Portfolio"
            ))

            # custom portfolios
            ret_custom, vol_custom, stocks_custom, wts_custom = [], [], [], []

            if len(custom_weights) > 0:
                for wts in custom_weights:
                    wts = np.array(wts)
                    ret_custom.append(np.sum(wts * rets))
                    vol_custom.append(np.sqrt(np.dot(wts.T, np.dot(cov, wts))))
                    stocks_custom.append(list(df.columns))
                    wts_custom.append(wts)

            if len(custom_weights) > 0:
                fig.add_trace(go.Scatter(
                    y = ret_custom,
                    x = vol_custom,
                    mode = 'markers',
                    name = 'Custom Portfolios',
                    marker = dict(color='blue', size=16),
                    marker_symbol = 'triangle-up',
                    text = [str(np.array(stocks_custom[i])) + "<br>" + str(np.array(wts_custom[i]).round(2)) for i in range(len(stocks_custom))]
                ))

            # Add title/labels
            fig.update_layout(template='plotly_white',
                            xaxis=dict(title='Annualised Risk (Volatility)'),
                            yaxis=dict(title='Annualised Return'),
                            title='Markowitz Efficient Frontier',
                            coloraxis_colorbar=dict(title="Sharpe Ratio"),
                            legend=dict(yanchor="bottom", y=0.1, xanchor="right", x=0.99),
                            font=dict(family = "Courier New, monospace",
                                      size = 12,
                                      color="RebeccaPurple"
                                    ), width = graph_width, height = graph_height)

            return fig

    # Exceptions other than Assertion Errors.
    except Exception as e:
        print(e)





def return_plots(df:pd.DataFrame, frequency:str="D", strategy:str='', graph_width:int=1600, graph_height:int=600, yaxis:str="Cumulative Return (%)",
                                                                                                        xaxis:str="", title:str="Cumulative Returns", x_shift:int=25) -> go.Figure:
    '''
    Plots the returns over a period of time for all assets in the given
    dataframe.

    Parameters
    ----------
    df : Pandas DataFrame object
        DataFrame containing the asset prices
    frequency : str, default 'D'    Frequency of the data
    strategy : str, default - blank string
        Name of the strategy
    graph_width : int, default - 1600
        Width of the graph object
    graph_height : int, default - 600
        Height of the graph object
    yaxis : str, default - 'Cumulative Return (%)'
        Label on the yaxis
    xaxis : str, default - ''
        Label on the xaxis
    title : str, default - ''
        Title of the chart
    x_shift : int, default - 25
        Position of the annotation

    Returns
    -------
    plotly.graph_objects : Plots the Return Charts

    Example
    -------
    >>> df.head(2)

                    TCS  NTPC  INFY
    Dates
    2022-01-01      200  300   400
    2022-01-02      202  299   401


    >>> return_plots(df)
    '''
    # Type checking
    assert (str(type(df)) == "<class 'pandas.core.frame.DataFrame'>"), "Not a dataframe"
    assert (frequency in ['D', 'M', 'Q', 'W', 'Y', '6M']), "Should be one of the values of 'D', 'M', 'Q', 'W', 'Y', '6M'"
    assert (str(type(graph_width)) == "<class 'int'>"), "Not an integer value"
    assert (str(type(graph_height)) == "<class 'int'>"), "Not an integer value"
    assert (str(type(strategy)) == "<class 'str'>"), "Not a string"
    assert (str(type(yaxis)) == "<class 'str'>"), "Not a string"
    assert (str(type(xaxis)) == "<class 'str'>"), "Not a string"
    assert (str(type(title)) == "<class 'str'>"), "Not a string"

    try:
        # Cumulative Returns
        cum_returns = ((1 + df.pct_change().fillna(0)).cumprod() - 1) * 100

        # Mapping the frequencies
        mapper = {'D': 'Day', 'M': 'Month', 'W': 'Week', 'Q': 'Quarter', 'Y': 'Year', '6M': 'Half Yearly'}

        if xaxis == "":
            xaxis = mapper[frequency]

        # Plotting all plots
        fig = go.Figure()

        for assets in cum_returns.columns:
            fig.add_trace(go.Scatter(
                y = cum_returns[f'{assets}'],
                x = cum_returns.index,
                name = f'{assets}'
            ))

            fig.add_annotation(
                y=cum_returns[f'{assets}'].iloc[-1],
                x=cum_returns.index[-1],
                text=str(round(cum_returns[f'{assets}'].iloc[-1], 2)),
                showarrow=False,
                xshift=x_shift,
                bgcolor='rgb(255,0,0)',
                font_color='rgb(255,255,255)'
            )

        fig.update_layout(yaxis_title=yaxis, xaxis_title=xaxis, title=f"{title} {strategy}", font=dict(family = "Courier New, monospace",
                                      size = 12,
                                      color="RebeccaPurple"
                                    ), width = graph_width, height = graph_height)

        return fig

    except Exception as e:
        print(e)





def rolling_volatility_plots(df:pd.DataFrame, frequency:str="D", windows:int=30, strategy:str='', graph_width:int=1600, graph_height:int=600, yaxis:str="Volatility (%)",
                                                                                                        xaxis:str="", title:str="Rolling Volatility", x_shift:int=25) -> go.Figure:
    '''
    Plots the rolling volatilities over a period of time for all assets in the given
    dataframe.

    Parameters
    ----------
    df : Pandas DataFrame object
        DataFrame containing the asset prices
    frequency : str, default 'D'    Frequency of the data
    windows : int, default 30
        Size of the rolling window for calculation of volatility
    strategy : str, default - blank string
        Name of the strategy
    graph_width : int, default - 1600
        Width of the graph object
    graph_height : int, default - 600
        Height of the graph object
    yaxis : str, default - 'Cumulative Return (%)'
        Label on the yaxis
    xaxis : str, default - ''
        Label on the xaxis
    title : str, default - ''
        Title of the chart
    x_shift : int or float, default - 25
        Position of annotation

    Returns
    -------
    plotly.graph_objects : Plots the Rolling Volatility Charts

    Example
    -------
    >>> df.head(2)

                    TCS  NTPC  INFY
    Dates
    2022-01-01      200  300   400
    2022-01-02      202  299   401


    >>> volatility_plots(df)
    '''
    # Type checking
    assert (str(type(df)) == "<class 'pandas.core.frame.DataFrame'>"), "Not a dataframe"
    assert (frequency in ['D', 'M', 'Q', 'W', 'Y', '6M']), "Should be one of the values of 'D', 'M', 'Q', 'W', 'Y', '6M'"
    assert (str(type(windows))) == "<class 'int'>", "Expects an integer value"
    assert (str(type(graph_width)) == "<class 'int'>"), "Not an integer value"
    assert (str(type(graph_height)) == "<class 'int'>"), "Not an integer value"
    assert (str(type(strategy)) == "<class 'str'>"), "Not a string"
    assert (str(type(yaxis)) == "<class 'str'>"), "Not a string"
    assert (str(type(xaxis)) == "<class 'str'>"), "Not a string"
    assert (str(type(title)) == "<class 'str'>"), "Not a string"


    try:
        # Mapping the frequencies
        mapper = {'D': 'Day', 'M': 'Month', 'W': 'Week', 'Q': 'Quarter', 'Y': 'Year', '6M': 'Half Yearly'}

        # Time Mapper
        mapper_time = {'D': 252, 'M': 12, 'W': 52, 'Q': 4, 'Y': 1, '6M': 2}

        # Volatility
        volas = df.pct_change().rolling(window=windows).std().apply(lambda x: x * np.sqrt(mapper_time[frequency])).dropna() * 100

        if xaxis == "":
            xaxis = mapper[frequency]

        # Plotting all plots
        fig = go.Figure()

        for assets in volas.columns:
            fig.add_trace(go.Scatter(
                y = volas[f'{assets}'],
                x = volas.index,
                name = f'{assets}'
            ))

            fig.add_annotation(
                y=volas[f'{assets}'].iloc[-1],
                x=volas.index[-1],
                text=str(round(volas[f'{assets}'].iloc[-1], 2)),
                showarrow=False,
                xshift=x_shift,
                bgcolor='rgb(255,0,0)',
                font_color='rgb(255,255,255)'
            )

        fig.update_layout(yaxis_title=yaxis, xaxis_title=xaxis, title=f"{title} {strategy}", font=dict(family = "Courier New, monospace",
                                      size = 12,
                                      color="RebeccaPurple"
                                    ), width = graph_width, height = graph_height)
        return fig

    except Exception as e:
        print(e)





def rolling_max_drawdown_plots(df:pd.DataFrame, frequency:str='D', windows:int=10, strategy:str='', graph_width:int=1600, graph_height:int=600,
                                                        yaxis:str="Maximum Drawdown in (%)", xaxis:str="", title:str="Rolling Maximum Drawdown") -> go.Figure:
    '''
    Plots the rolling max drawdown over a period of time for all assets in the given
    dataframe.

    Parameters
    ----------
    df : Pandas DataFrame object
        DataFrame containing the asset prices
    frequency : str, default 'D'    Frequency of the data
    windows : int, default 10
        Size of the rolling window for calculation of max drawdown
    strategy : str, default - blank string
        Name of the strategy
    graph_width : int, default - 1600
        Width of the graph object
    graph_height : int, default - 600
        Height of the graph object
    yaxis : str, default - 'Cumulative Return (%)'
        Label on the yaxis
    xaxis : str, default - ''
        Label on the xaxis
    title : str, default - ''
        Title of the chart

    Returns
    -------
    plotly.graph_objects : Plots the Rolling Maximum Drawdown Charts

    Example
    -------
    >>> df.head(2)

                    TCS  NTPC  INFY
    Dates
    2022-01-01      200  300   400
    2022-01-02      202  299   401


    >>> drawdown_plots(df)
    '''
    # Type checking
    assert (str(type(df)) == "<class 'pandas.core.frame.DataFrame'>"), "Not a dataframe"
    assert (frequency in ['D', 'M', 'Q', 'W', 'Y', '6M']), "Should be one of the values of 'D', 'M', 'Q', 'W', 'Y', '6M'"
    assert (str(type(windows))) == "<class 'int'>", "Expects an integer value"
    assert (str(type(graph_width)) == "<class 'int'>"), "Not an integer value"
    assert (str(type(graph_height)) == "<class 'int'>"), "Not an integer value"
    assert (str(type(strategy)) == "<class 'str'>"), "Not a string"
    assert (str(type(yaxis)) == "<class 'str'>"), "Not a string"
    assert (str(type(xaxis)) == "<class 'str'>"), "Not a string"
    assert (str(type(title)) == "<class 'str'>"), "Not a string"


    try:
        # Mapping the frequencies
        mapper = {'D': 'Day', 'M': 'Month', 'W': 'Week', 'Q': 'Quarter', 'Y': 'Year', '6M': 'Half Year'}

        # Maximum Drawdown
        rolling_max = df.rolling(window=windows).max()
        rolling_max_diff_close = (df / rolling_max) - 1
        max_drawdown_ = rolling_max_diff_close.rolling(window=windows).min().dropna() * 100

        if xaxis == "":
            xaxis = mapper[frequency]


        # Plotting all plots
        fig = go.Figure()

        for assets in max_drawdown_.columns:
            fig.add_trace(go.Scatter(
                y = max_drawdown_[f'{assets}'],
                x = max_drawdown_.index,
                name = f'{assets}',
                fill = 'tozeroy'
            ))

        fig.update_layout(yaxis_title=yaxis, xaxis_title=xaxis, title=f"{title} {strategy}", font=dict(family = "Courier New, monospace",
                                      size = 12,
                                      color = "RebeccaPurple"
                                    ), width = graph_width, height = graph_height)
        return fig

    except Exception as e:
        print(e)





def annualised_periodic_returns(df:pd.DataFrame, securities:list, strategy:str='', frequency:str='D', graph_width:int=1600, graph_height:int=600,
                                                                    yaxis:str="Return Percentage (%)", xaxis:str="", title:str="Annualised Return") -> go.Figure:
    '''
    Plots the annualised returns over a period of time in the given dataframe

    Parameters
    ----------
    df : Pandas DataFrame object
        DataFrame containing the asset prices
    frequency : str, default 'D'   Frequency of the data
    securities : list, default - blank list
        Contains the name of the securities
    strategy : str, default - blank string
        Name of the strategy
    graph_width : int, default - 1600
        Width of the graph object
    graph_height : int, default - 600
        Height of the graph object
    yaxis : str, default - 'Cumulative Return (%)'
        Label on the yaxis
    xaxis : str, default - ''
        Label on the xaxis
    title : str, default - ''
        Title of the chart

    Returns
    -------
    plotly.graph_objects : Plots the Annualised Returns

    '''

    # Type checking
    assert (str(type(df)) == "<class 'pandas.core.frame.DataFrame'>"), "Not a dataframe"
    assert (frequency in ['D', 'M', 'Q', 'W', 'Y', '6M']), "Should be one of the values of 'D', 'M', 'Q', 'W', 'Y', '6M'"
    assert (str(type(graph_width)) == "<class 'int'>"), "Not an integer value"
    assert (str(type(graph_height)) == "<class 'int'>"), "Not an integer value"
    assert (str(type(strategy)) == "<class 'str'>"), "Not a string"
    assert (str(type(securities)) == "<class 'list'>"), "Not a list object"
    assert (str(type(yaxis)) == "<class 'str'>"), "Not a string"
    assert (str(type(xaxis)) == "<class 'str'>"), "Not a string"
    assert (str(type(title)) == "<class 'str'>"), "Not a string"


    try:
        # Mapping the frequencies
        mapper = {'D': 'Day', 'M': 'Month', 'W': 'Week', 'Q': 'Quarter', 'Y': 'Year', '6M': 'Half Yearly'}

        if xaxis == "":
            xaxis = mapper[frequency]

        fig = go.Figure()

        # plotting each security and their means
        for name in securities:
            periods = list(df.columns)
            periods.remove('Security Name')

            periodic_returns = (df[df['Security Name']==name].loc['Return'].values[:]).tolist()
            periodic_returns.remove(name)
            periodic_returns = np.array(periodic_returns).astype('float')
            mean_returns = np.mean(periodic_returns)


            fig.add_trace(go.Bar(
                y = periodic_returns,
                x = periods,
                name = f'{name}',
                text = periodic_returns,
                textposition = 'outside'
            ))


            fig.add_trace(go.Scatter(y = [mean_returns, mean_returns],
                                    x = [min(periods), max(periods)],
                                    mode = 'lines',
                                    line = dict(width=2, dash='dash'),
                                    name = f'Mean {name}')
                        )

        fig.update_layout(yaxis_title=f"{yaxis}", xaxis_title=f"{xaxis}", showlegend=True, title=f"{title} {strategy}", font=dict(
                                                                                                                                    family = "Courier New, monospace",
                                                                                                                                    size = 12,
                                                                                                                                    color="RebeccaPurple"
                                                                                                                                    ), width = graph_width, height = graph_height)
        return fig

    except Exception as e:
        print(e)





def max_drawdown_plots(df:pd.DataFrame, strategy:str='', frequency:str='D', graph_width:int=1600, graph_height:int=600, yaxis:str="Maximum Drawdown (%)",
                                                                                                                    xaxis:str="", title:str="Maximum Drawdown") -> go.Figure:
    '''
    Plots the drawdown over a period of time for all assets in the given
    dataframe.

    Parameters
    ----------
    df : Pandas DataFrame object
        DataFrame containing the asset prices
    frequency : str, default 'D'   Frequency of the data
    strategy : str, default - blank string
        Name of the strategy
    graph_width : int, default - 1600
        Width of the graph object
    graph_height : int, default - 600
        Height of the graph object
    yaxis : str, default - 'Cumulative Return (%)'
        Label on the yaxis
    xaxis : str, default - ''
        Label on the xaxis
    title : str, default - ''
        Title of the chart

    Returns
    -------
    plotly.graph_objects : Plots the Drawdown Charts

    Example
    -------
    >>> df.head(2)
                    TCS  NTPC  INFY
    Dates
    2022-01-01      200  300   400
    2022-01-02      202  299   401

    >>> drawdown_plots(df)
    '''
    # Type checking
    assert (str(type(df)) == "<class 'pandas.core.frame.DataFrame'>"), "Not a dataframe"
    assert (frequency in ['D', 'M', 'Q', 'W', 'Y', '6M']), "Should be one of the values of 'D', 'M', 'Q', 'W', 'Y', '6M'"
    assert (str(type(graph_width)) == "<class 'int'>"), "Not an integer value"
    assert (str(type(graph_height)) == "<class 'int'>"), "Not an integer value"
    assert (str(type(strategy)) == "<class 'str'>"), "Not a string"
    assert (str(type(yaxis)) == "<class 'str'>"), "Not a string"
    assert (str(type(xaxis)) == "<class 'str'>"), "Not a string"
    assert (str(type(title)) == "<class 'str'>"), "Not a string"

    try:
        # Mapping the frequencies
        mapper = {'D': 'Day', 'M': 'Month', 'W': 'Week', 'Q': 'Quarter', 'Y': 'Year', '6M': 'Half Yearly'}

        if xaxis == "":
            xaxis = mapper[frequency]

        returns = df.pct_change()
        df_cum_rets = ep.cum_returns(returns, starting_value=1.0)
        running_max = np.maximum.accumulate(df_cum_rets)
        underwater = -100 * ((running_max - df_cum_rets) / running_max)

        # Plotting all plots
        fig = go.Figure()
        for assets in underwater.columns:
            fig.add_trace(go.Scatter(
                y = underwater[f'{assets}'],
                x = underwater.index,
                fill='tozeroy',
                name = f'{assets}'
            ))

        fig.update_layout(yaxis_title=yaxis, xaxis_title=f"{xaxis}", title=f"{title} {strategy}", font=dict(family = "Courier New, monospace",
                                                                                                            size = 12,
                                                                                                            color = "RebeccaPurple"
                                                                                                        ), width = graph_width, height = graph_height)

        return fig

    except Exception as e:
        print(e)





def plotly_create_subplots(*figs, column: int, row: int, subplot_title: tuple,
            xaxes_title=None, yaxes_title=None, hs:float=0.15) -> go.Figure:
    """
    This function helps us to create sub plots in plotly

    Parameters:
    -----------
        column : int
            Number of columns
        row : int
            Number of rows
        subplot_title : tuple
            subplot title
        xaxes_title : str, optional
            Title of X axis (defaults None).
        yaxes_title : str, optional
            Title of Y axis (defaults None).
        hs : float, default - 0.15
            Horizontal spacing between the subplots

    Returns:
    --------
        go.Figure : object
    """

    assert(len(figs) <= column*row), "Shape mismatched"
    assert(len(figs) == len(subplot_title)), "Number of subplot and number of subplot title are mismatched"

    try:
        fig_sub_plot_object = make_subplots(rows=row, cols=column, subplot_titles=subplot_title, horizontal_spacing=hs)
        row_idx = 1
        col_idx = 1
        for fig in figs:
            for i in range(len(fig.data)):
                fig_sub_plot_object.add_trace(fig.data[i], row=row_idx, col=col_idx)
            fig_sub_plot_object.update_xaxes(title_text=xaxes_title, row=row_idx, col=col_idx)
            fig_sub_plot_object.update_yaxes(title_text=yaxes_title, row=row_idx, col=col_idx)
            col_idx += 1
            if col_idx > column:
                row_idx += 1
                col_idx = 1

        return fig_sub_plot_object

    except Exception as e:
        print(e)





def plotly_figure_show(figure: go.Figure, height: int=600, width: int=800, figure_title: str=None) -> None:
    """
    This function help us to show the plotly plots

    Parameters:
    -----------
        figure : plotly.Figure
            plotly figure object which we want to show
        height : int, optional
            height of the plot (default 600).
        width : int, optional
            width of the plot (default 800).
        figure_title : str, optional
            title of the figure. (defaults None).
    """
    try:

        figure.update_layout(height=height, width=width, title_text=figure_title)
        figure.show()

    except Exception as e:
        print(e)





def general_plots(df:pd.DataFrame, frequency:str="D", strategy:str='', graph_width:int=1600, graph_height:int=600, y_title:str='', chart_title:str='', x_title:str='', x_shift:int=25) -> go.Figure:
    '''
    Plots any given dataframe

    Parameters
    ----------
    df : Pandas DataFrame object
        DataFrame containing the asset prices
    frequency : str, default 'D'    Frequency of the data
    strategy : str, default - blank string
        Name of the strategy
    graph_width : int, default - 1600
        Width of the graph object
    graph_height : int, default - 600
        Height of the graph object
    y_title : str, default - ''
        Y axis title
    x_title : str, default - ''
        X axis title
    chart_title : str, default - ''
        Title of the chart
    x_shift : int, default - 25
        position of annotation on axis

    Returns
    -------
    plotly.graph_objects : Plots Any DataFrame

    Example
    -------
    >>> df.head(2)

                    TCS  NTPC  INFY
    Dates
    2022-01-01      200  300   400
    2022-01-02      202  299   401


    >>> general_plots(df)
    '''
    # Type checking
    assert (str(type(df)) == "<class 'pandas.core.frame.DataFrame'>"), "Not a dataframe"
    assert (frequency in ['D', 'M', 'Q', 'W', 'Y', '6M']), "Should be one of the values of 'D', 'M', 'Q', 'W', 'Y', '6M'"
    assert (str(type(graph_width)) == "<class 'int'>"), "Not an integer value"
    assert (str(type(graph_height)) == "<class 'int'>"), "Not an integer value"
    assert (str(type(strategy)) == "<class 'str'>"), "Not a string"
    assert (str(type(y_title)) == "<class 'str'>"), "Not a string"
    assert (str(type(chart_title)) == "<class 'str'>"), "Not a string"
    assert (str(type(x_title)) == "<class 'str'>"), "Not a string"

    try:
        # Mapping the frequencies
        mapper = {'D': 'Day', 'M': 'Month', 'W': 'Week', 'Q': 'Quarter', 'Y': 'Year', '6M': 'Half Yearly'}

        # Plotting all plots
        fig = go.Figure()

        for assets in df.columns:
            fig.add_trace(go.Scatter(
                y = df[f'{assets}'],
                x = df.index,
                name = f'{assets}'
            ))

            fig.add_annotation(
                y=df[f'{assets}'].iloc[-1],
                x=df.index[-1],
                text=str(round(df[f'{assets}'].iloc[-1], 2)),
                showarrow=False,
                xshift=x_shift,
                bgcolor='rgb(255,0,0)',
                font_color='rgb(255,255,255)'
            )

        fig.update_layout(yaxis_title=f"{y_title}", xaxis_title=f"{x_title}", title=f"{chart_title} {strategy}", font=dict(family = "Courier New, monospace",
                                      size = 12,
                                      color="RebeccaPurple"
                                    ), width = graph_width, height = graph_height)
        return fig

    except Exception as e:
        print(e)





def plot_returns_heatmap(returns:pd.DataFrame, z:float, strategy:str='', frequency:str='D', graph_width:int=1600, graph_height:int=600, color_scale_heatmap:str='RdYlGn',
                                                                                                    yaxis:str="Month",
                                                                                                    xaxis:str="", title:str="Return Heatmap in (%)", colorbar_title:str="Return (%)") -> go.Figure:
    """
    Plots a heatmap of returns

    Parameters
    ----------
    returns : pd.DataFrame
        Daily, Monthly, Weekly etc. returns of the strategy, noncumulative
    frequency : str, default 'D'    Frequency of the data
    strategy : str, default - blank string
        Name of the strategy
    graph_width : int, default - 1600
        Width of the graph object
    graph_height : int, default - 600
        Height of the graph object
    color_scale_heatmap : str, default - "RdYlGn"
        color scale of the heatmap
    yaxis : str, default - 'Cumulative Return (%)'
        Label on the yaxis
    xaxis : str, default - ''
        Label on the xaxis
    title : str, default - ''
        Title of the chart
    colorbar_title : str, default - Return
        Title of the colorbar
    z : float, int default - None
        range of the colorbar

    Returns
    -------
    plotly.graph_objects : Plots a Heatmap
    """
    # Type checking
    assert (str(type(returns)) == "<class 'pandas.core.frame.DataFrame'>"), "Not a dataframe"
    assert (frequency in ['D', 'M', 'Q', 'W', 'Y', '6M']), "Should be one of the values of 'D', 'M', 'Q', 'W', 'Y', '6M'"
    assert (str(type(graph_width)) == "<class 'int'>"), "Not an integer value"
    assert (str(type(graph_height)) == "<class 'int'>"), "Not an integer value"
    assert (str(type(strategy)) == "<class 'str'>"), "Not a string"
    assert (str(type(color_scale_heatmap)) == "<class 'str'>"), "Not a string"
    assert (str(type(yaxis)) == "<class 'str'>"), "Not a string"
    assert (str(type(xaxis)) == "<class 'str'>"), "Not a string"
    assert (str(type(title)) == "<class 'str'>"), "Not a string"
    assert (str(type(colorbar_title)) == "<class 'str'>"), "Not a string"

    try:
        # Mapping the frequencies
        mapper = {'D': 'Day', 'M': 'Month', 'W': 'Week', 'Q': 'Quarter', 'Y': 'Year', '6M': 'Half Yearly'}

        if xaxis == "":
            xaxis = "Year"

        fig = px.imshow(returns,
                         labels=dict(color=f"{colorbar_title}"),
                         color_continuous_scale=color_scale_heatmap,
                         text_auto=True,
                         zmin=-z,
                         zmax=z
               )

        fig.update_layout(yaxis_title=f"{yaxis}", xaxis_title=f"{xaxis}", title=f"{title} {strategy}", font=dict(family = "Courier New, monospace",
                                          size = 12,
                                          color="RebeccaPurple"
                                        ), width = graph_width, height = graph_height)


        return fig

    except Exception as e:
        print(e)





def rv_distribution_scatter_plots(data:pd.DataFrame, scatter:bool=False, colors:list=['#A56CC1'], graph_width:int=800, graph_height:int=400, frequency:str='M',
                                                            price_col:str='Adj Close', date_col:str='Date', single_scatter=False, annualised:bool=False) -> go.Figure:
    '''
    Plots the scatter and distribution plots for return and volatilities

    Parameters
    ----------
    data : pd.DataFrame, default - None
        A dataframe containing the prices with a Date column
    scatter : bool, default - False
        Plots the scatter plot
    single_scatter : bool, deafult - False
        Plot single variable scatter plots
    colors: list, default - ['#A56CC1']
        colors for the plots
    graph_width : int, default - 1600
        Width of the graph object
    graph_height : int, default - 600
        Height of the graph object
    frequency : str, default - 'M'
        frequncy over which Return and Volatility plots to be plotted
    price_col : str, default - 'Adj Close'
        Price column of the dataframe
    annualised : bool, default - False
        Annualise the returns and the volatility
    date_col : str, default - 'Date
        Name of the date column
    Return
    ------
    plotly.graph_objects (tuples): Plotly scatter and displot objects

    3 object tuple : single_scatter = True
    1 object tuple : single_scatter = False, scatter = True
    2 object tuple : single_scatter = False, scatter = False
    '''
    assert (str(type(data)) == "<class 'pandas.core.frame.DataFrame'>"), "Not a dataframe"
    assert (str(type(scatter)) == "<class 'bool'>"), "Not a boolean value"
    assert (str(type(single_scatter)) == "<class 'bool'>"), "Not a boolean value"
    assert (str(type(annualised)) == "<class 'bool'>"), "Not a boolean value"
    assert (str(type(graph_width)) == "<class 'int'>"), "Not an integer value"
    assert (str(type(graph_height)) == "<class 'int'>"), "Not an integer value"
    assert (frequency in ['M', 'Q', 'W', 'Y', '6M', 'D']), "Should be one of the values of 'M', 'Q', 'W', 'Y', '6M', 'D'"
    assert (str(type(price_col)) == "<class 'str'>"), "Not a string"
    assert (str(type(date_col)) == "<class 'str'>"), "Not a string"

    try:
        data[date_col] = pd.to_datetime(data[date_col])
        data.set_index(date_col, inplace=True)
        trading_days = - ({'M': 22, 'W': 5, 'Q': 66, 'D': 1, '6M': 126, 'Y': 252}[frequency])

        # Time Mapper
        mapper_time = {'D': 252, 'M': 12, 'W': 52, 'Q': 4, 'Y': 1, '6M': 2}

        periodic = data.resample(frequency).last()
        ret, vol = [], []
        prev = periodic.index[0]

        # Annualised
        for dates in periodic.index.unique()[1:]:
            if annualised:

                vol_cal = np.std(data.loc[:dates, price_col][trading_days:].pct_change().dropna().values) * np.sqrt(mapper_time[frequency])
                vol.append(vol_cal)
                rets = ((periodic.loc[dates][price_col] / periodic.loc[prev][price_col]) - 1) * mapper_time[frequency]
                ret.append(rets)
                prev = dates
            else:

                vol_cal = np.std(data.loc[:dates, price_col][trading_days:].pct_change().dropna().values)
                vol.append(vol_cal)
                rets = ((periodic.loc[dates][price_col] / periodic.loc[prev][price_col]) - 1)
                ret.append(rets)
                prev = dates

        # percentage format
        ret = pd.Series(ret).dropna().apply(lambda x: x * 100).values.tolist()
        vol = pd.Series(vol).dropna().apply(lambda x: x * 100).values.tolist()

        if not scatter:

            hist_data = [ret]

            group_labels = ['Return']
            colors = colors

            fig = ff.create_distplot(hist_data, group_labels, colors=colors,
                                    bin_size=.2, show_rug=False)

            # Return Distribution
            fig.update_layout(title_text='Return Distribution', xaxis_title="Return (%)", yaxis_title="Probability Density")



            group_labels = ['Volatility']
            hist_data = [vol]

            fig1 = ff.create_distplot(hist_data, group_labels, colors=colors,
                                    bin_size=.2, show_rug=False)

            # Volatility Distribution
            fig1.update_layout(title_text='Volatility Distribution', xaxis_title="Volatility (%)", yaxis_title="Probability Density", font=dict(family = "Courier New, monospace",
                                          size = 12,
                                          color="RebeccaPurple"
                                        ), width = graph_width, height = graph_height)

            return fig, fig1

        else:

            if single_scatter:
                # Single Volatility
                fig2 = go.Figure()

                fig2.add_trace(go.Scatter(
                    x = vol,
                    y = [0]*len(vol),
                    mode = 'markers',
                    marker = dict(size=7, color = (
                            pd.Series(ret) > 0
                        ).astype('int'),
                    colorscale = ([[0, 'red'], [1, 'green']])),
                    name = 'Volatility'
                ))

                fig2.update_layout(
                    title = "Volatility Scatter Plot",
                    xaxis_title = "Volatility (%)",
                    yaxis_visible = False,
                    font=dict(family = "Courier New, monospace",
                                              size = 12,
                                              color="RebeccaPurple"
                                            ), width = graph_width, height = graph_height
                )

                # Single Return
                fig3 = go.Figure()

                fig3.add_trace(go.Scatter(
                    x = ret,
                    y = [0]*len(ret),
                    mode = 'markers',
                    marker = dict(size=7, color = (
                            pd.Series(ret) > 0
                        ).astype('int'),
                    colorscale = ([[0, 'red'], [1, 'green']])),
                    name = 'Return'
                ))

                fig3.update_layout(
                    title = "Return Scatter Plot",
                    xaxis_title = "Return (%)",
                    yaxis_visible = False,
                    font=dict(family = "Courier New, monospace",
                                              size = 12,
                                              color="RebeccaPurple"
                                            ), width = graph_width, height = graph_height
                )



            fig = go.Figure()

            fig.add_trace(go.Scatter(
                x = vol,
                y = ret,
                mode = 'markers',
                name = 'Return vs Volatility',
                marker = dict(
                    size = 7,
                    color = (
                            pd.Series(ret) > 0
                        ).astype('int'),
                    colorscale=[[0, 'red'], [1, 'green']]
                    )
                )
            )

            fig.update_layout(
                title = "Return-Volatility",
                xaxis_title = "Volatility (%)",
                yaxis_title = "Return (%)",
                font=dict(family = "Courier New, monospace",
                                          size = 12,
                                          color="RebeccaPurple"
                                        ), width = graph_width, height = graph_height
            )

            if single_scatter:

                return fig, fig2, fig3

            return fig

    except Exception as e:
        print(e)





def rv_time_distribution_scatter_plots(data:pd.DataFrame, colors:list=['#A56CC1'], graph_width:int=800, graph_height:int=400, frequency:str='M',
                                                                                         price_col:str='Adj Close', date_col:str='Date', annualised:bool=False) -> go.Figure:
    '''
    Plots the time distribution plots for return and volatilities.

    Parameters
    ----------
    data : pd.DataFrame, default - None
        A dataframe containing the prices with a Date column
    colors: list, default - ['#A56CC1']
        colors for the plots
    graph_width : int, default - 1600
        Width of the graph object
    graph_height : int, default - 600
        Height of the graph object
    frequency : str, default - 'M'
        frequncy over which Return and Volatility plots to be plotted
    price_col : str, default - 'Adj Close'
        Price column of the dataframe
    annualised : bool, default - False
        Annualise the returns and the volatility
    date_col : str, default - 'Date
        Name of the date column

    Return
    ------
    plotly.graph_objects (tuples): Plotly scatter and displot objects

    2 Figure objects (tuple)
    '''
    assert (str(type(data)) == "<class 'pandas.core.frame.DataFrame'>"), "Not a dataframe"
    assert (str(type(graph_width)) == "<class 'int'>"), "Not an integer value"
    assert (str(type(graph_height)) == "<class 'int'>"), "Not an integer value"
    assert (frequency in ['D', 'M', 'Q', 'W', 'Y', '6M']), "Should be one of the values of 'M', 'Q', 'W', 'Y', '6M', 'D'"
    assert (str(type(price_col)) == "<class 'str'>"), "Not a string"
    assert (str(type(date_col)) == "<class 'str'>"), "Not a string"
    assert (str(type(annualised)) == "<class 'bool'>"), "Not a boolean value"

    try:
        data[date_col] = pd.to_datetime(data[date_col])
        data.set_index(date_col, inplace=True)
        trading_days = - ({'M': 22, 'W': 5, 'Q': 66, 'D': 1, '6M': 126, 'Y': 252}[frequency])

        # Time Mapper
        mapper_time = {'D': 252, 'M': 12, 'W': 52, 'Q': 4, 'Y': 1, '6M': 2}

        periodic = data.resample(frequency).last()
        ret, vol = [], []
        prev = periodic.index[0]

        # Annualised
        for dates in periodic.index.unique()[1:]:
            if annualised:

                vol_cal = np.std(data.loc[:dates, price_col][trading_days:].pct_change().dropna().values) * np.sqrt(mapper_time[frequency])
                vol.append(vol_cal)
                rets = ((periodic.loc[dates][price_col] / periodic.loc[prev][price_col]) - 1) * mapper_time[frequency]
                ret.append(rets)
                prev = dates
            else:

                vol_cal = np.std(data.loc[:dates, price_col][trading_days:].pct_change().dropna().values)
                vol.append(vol_cal)
                rets = ((periodic.loc[dates][price_col] / periodic.loc[prev][price_col]) - 1)
                ret.append(rets)
                prev = dates

        # percentage format
        ret = pd.Series(ret).dropna().apply(lambda x: x * 100).values.tolist()
        vol = pd.Series(vol).dropna().apply(lambda x: x * 100).values.tolist()



        fig = go.Figure()

        fig.add_trace(go.Scatter(
            x = periodic.index,
            y = ret,
            mode = 'markers',
            name = 'Return vs Volatility',
            marker = dict(
                size = 7,
                color = (
                        pd.Series(ret) > 0
                    ).astype('int'),
                colorscale=[[0, 'red'], [1, 'green']]
                )
            )
        )

        fig.update_layout(
            title = "Time Distribution of Returns",
            xaxis_title = "Time",
            yaxis_title = "Return (%)",
            font=dict(family = "Courier New, monospace",
                                      size = 12,
                                      color="RebeccaPurple"
                                    ), width = graph_width, height = graph_height
        )



        fig1 = go.Figure()

        fig1.add_trace(go.Scatter(
            x = periodic.index,
            y = vol,
            mode = 'markers',
            name = 'Return vs Volatility',
            marker = dict(
                size = 7,
                color = (
                        pd.Series(ret) > 0
                    ).astype('int'),
                colorscale=[[0, 'red'], [1, 'green']]
                )
            )
        )

        fig1.update_layout(
            title = "Time Distribution of Volatility",
            xaxis_title = "Time",
            yaxis_title = "Volatility (%)",
            font=dict(family = "Courier New, monospace",
                                      size = 12,
                                      color="RebeccaPurple"
                                    ), width = graph_width, height = graph_height
        )


        return fig, fig1

    except Exception as e:
        print(e)





def candlestick_plot(
        df : pd.DataFrame,
        date_column_name : str = 'Date',
        chart_title : str = 'Candlestick Chart',
        xaxis_title : str = 'Date',
        yaxis_title : str = 'Prices',
        xaxis_rangeslider_visible : bool = False,
        graph_width:int=800,
        graph_height:int=400,
        yaxis_side : str = 'right',
        xshift_value : int = 25,
        yshift_value : int = 0
    ) -> go.Figure:
    '''
    Plots the Candlestick chart of stock data.

    Parameters
    ----------
    df : pd.DataFrame, default - None
        A dataframe containing the prices with a Date column (must have Open, High, Low, Close columns)
    date_column_name: str, default - 'Date'
        Date column of the dataset
    graph_width : int, default - 1600
        Width of the graph object
    graph_height : int, default - 600
        Height of the graph object
    yaxis_title: str, default - 'Prices'
        Title of the yaxis
    xaxis_title : str, default - 'Date'
        Title of the x-axis
    chart_title : str, default - 'Candlestick Chart'
        Name of the plot
    xaxis_rangeslider_visible : bool, default - False
        x-axis slider
    yaxis_side : str, default - 'right'
        Position of the y-axis
    xhift_value : int, default - 25
        Annotation to be shifted on the x-axis
    yshift_value : int, default - 0
        Annotation to be shifted on the yaxis

    Return
    ------
    plotly.graph_objects : Plotly object
    '''
    assert (str(type(df)) == "<class 'pandas.core.frame.DataFrame'>"), "Not a dataframe"
    assert (str(type(date_column_name)) == "<class 'str'>"), "Not a string"
    assert (str(type(chart_title)) == "<class 'str'>"), "Not a string"
    assert (str(type(xaxis_title)) == "<class 'str'>"), "Not a string"
    assert (str(type(yaxis_title)) == "<class 'str'>"), "Not a string"
    assert (str(type(yaxis_side)) == "<class 'str'>"), "Not a string"
    assert (str(type(xaxis_rangeslider_visible)) == "<class 'bool'>"), "Not a boolean value"
    assert (str(type(graph_width)) == "<class 'int'>"), "Not an integer value"
    assert (str(type(graph_height)) == "<class 'int'>"), "Not an integer value"
    assert (str(type(xshift_value)) == "<class 'int'>"), "Not an integer value"
    assert (str(type(yshift_value)) == "<class 'int'>"), "Not an integer value"

    try:
        df[date_column_name] = pd.to_datetime(df[date_column_name])
        df.set_index(date_column_name, inplace=True)

        # Create Chart Object

        fig = make_subplots(specs=[[{"secondary_y": True}]])

        fig.add_trace(

            go.Candlestick(

                x=df.index,

                open=df['Open'],

                high=df['High'],

                low=df['Low'],

                close=df['Close']

            ),

            secondary_y=True

        )

        fig.update_xaxes(title_text=xaxis_title, range=[df.index[0], df.index[-1]])

        fig.update_yaxes(title_text=yaxis_title, secondary_y=True)

        fig.update_layout(title=chart_title)



        fig.update_layout(

            width=graph_width,

            height=graph_height,

            xaxis_rangeslider_visible=xaxis_rangeslider_visible,

            yaxis=dict(side=yaxis_side)

        )



        # Determine text-annotation and back-ground color

        if df.Open[-1] > df.Close[-1] :

            y=df.Open[-1]

            bgcolor='rgb(3,75,3)' # deep-green



        if df.Open[-1] < df.Close[-1] :

            y=df.Close[-1]

            bgcolor='rgb(255,0,0)' # red



        # Show the annotation

        fig.add_annotation(

            x=df.index[-1],

            y=y,

            text=str(round(y)),

            showarrow=False,

            xshift=xshift_value,

            bgcolor=bgcolor,

            font_color='rgb(255,255,255)'

        )

        return fig

    except Exception as e:
        print(e)





def weight_graphs(weights_file:str, ticker:str, threshold:float, path:str='', price_col:str='Adj Close', frequency:str='M') -> go.Figure:
    '''
    Plots the weight charts and the buy/sell signals of individual assets.

    Parameters
    ----------
    weights_file:str, default-None
        A string containing the path of the weights of the asset (file to be .xlsx or a csv with sheet name as "Weights")
    ticker:str, default-None
        Name of the ticker - e.g (AAPL)
    threshold:float, default - None
        A value below which no bar is plotted
    path:str, default - ''
        A location where the raw prices of the asset is located. Leave it blank if inside the same folder where the function is being called.
        Eg: if asset prices is in a path such as 'Data/....csv' just give path='Data/'
    price_col:str, default-Adj Close
        Name of the price column
    frequency:str, deafult - 'M'
        Frequency of data supplied for weights

    Return
    ------
    go.Figure object: Weight and Buy/Sell Plots
    '''
    try:

        def get_options(df_weights:pd.DataFrame, ticker:str, threshold:float) -> np.array:
            df_weights = df_weights.apply(lambda x:round(x, 6))
            df_weights['difference'] = df_weights[ticker]-df_weights[ticker].shift(1)
            df_weights['difference_2'] = df_weights['difference'].where(np.absolute(df_weights['difference']) > threshold, other=0)
            options = np.sign(df_weights['difference_2'])
            df_weights.drop(['difference'], axis=1, inplace=True)

            if df_weights[ticker].iloc[0] > 0:
                options[0] = 1.0

            return options


        def df_function(weights_file:str, ticker:str, threshold:float, path:str):
            # Extracting the Ticker Weights
            try:
                df_weights = pd.read_csv(weights_file, index_col=0)
            except:
                df_weights = pd.read_excel(weights_file, index_col=0, sheet_name="Weights")

            df_weights.dropna(inplace=True)
            df_weights.index = pd.to_datetime(df_weights.index)
            df_ticker_weights = df_weights[[ticker]]
            df_ticker_weights.index = pd.Series(df_ticker_weights.index).apply(lambda x: x.strftime("%Y-%m-%d"))

            # Extracting the Adj Close for ticker
            ticker_csv_path = os.path.join(path,ticker+'.csv')
            df_price = pd.read_csv(ticker_csv_path, index_col=0)
            df_price.dropna(inplace=True)
            df_price.index = pd.to_datetime(df_price.index)
            df_ticker_price = df_price[[price_col]]
            df_ticker_price = df_ticker_price.resample(frequency).last()
            df_ticker_price.index = pd.Series(df_ticker_price.index).apply(lambda x: x.strftime("%Y-%m-%d"))
            df_ticker_price.index.name = 'date'

            # Merging Weights and Prices
            df = df_ticker_price.merge(df_ticker_weights, on='date')

            # Adding Option Column
            df['Option']=get_options(df_weights=df_ticker_weights, ticker=ticker, threshold=threshold)
            df.fillna(0, inplace=True)

            df[ticker] = df[ticker].where(df[ticker] > threshold, other=0)
            df[ticker] = df[ticker].apply(lambda x: x * 100)

            return df


        def get_bar_text(df:pd.DataFrame,ticker:str):
            text = []
            for i in df[ticker]:
                if round(i,2) == 0.00:
                    i = ''
                    text.append(i)
                else:
                    text.append(round(i*100, 2))
            text_df = pd.DataFrame(data={'text':text})
            text_df.replace(np.nan, '')

            return text_df.text

        # Extracting Weights and Prices
        df = df_function(weights_file, ticker=ticker, threshold=threshold, path=path)

        # Making figure object
        fig = make_subplots(
                        rows=2,
                        cols=1,
                        shared_xaxes=True,
                        vertical_spacing=0.02,
                        row_heights=[0.7,0.3],
                        specs=[
                                [{"secondary_y": True}],
                                [{"secondary_y": True}]
                            ]
                    )
        # Buy - markers
        fig.add_trace(
            go.Scatter(
                x=df[df['Option'] == 1].index,
                y=df[df['Option'] == 1][price_col],
                mode = 'markers',
                marker = dict(size = 10, symbol = 'triangle-up',color='green'),
                name='Buy'        ),
            row=1,
            col=1    )

        # Sell - markers
        fig.add_trace(
            go.Scatter(
                x=df[df['Option'] == -1].index,
                y=df[df['Option'] == -1][price_col],
                mode = 'markers',
                marker = dict(size = 10, symbol = 'triangle-down',color = 'red'),
                name='Sell'        ),
            row=1,
            col=1    )

        # Original Line (Line to be shown in plot)
        fig.add_trace(
            go.Scatter(
                x=df.index,
                y=df[price_col],
                mode='lines',
                name=f'{price_col}',
                marker = dict(color = 'orange')
            ),
            secondary_y=True,
            row=1,
            col=1    ).update_xaxes(
            range=[df.index[0],df.index[-1]]
        ).add_annotation(
                x=df.index[-1],
                y=df[price_col][-1],
                text=str(round(df[price_col][-1])),
                showarrow=False,
                xshift=15,
                bgcolor='rgb(255,0,0)',
                font_color='rgb(255,255,255)'        )

        # Pseudo Bar (Bar not to be shown in plot)
        fig.add_trace(
            go.Bar(
                x=df.index,
                y=df[ticker],
                showlegend=False,
                marker_color='white'        ),
            row=2,
            col=1    )

        df[ticker] = df[ticker].replace({0:np.nan})

        # Original Bar (Bar to be shown in plot)
        fig.add_trace(
            go.Bar(
                x=df.index,
                y=df[ticker],
                name='Weights',
                text=get_bar_text(df=df,ticker=ticker),
                textposition='outside',
                textfont_size=10,
                # width=0.1
            ),
            secondary_y=True,
            row=2,
            col=1    )

        # fig['layout']['yaxis']['title']='Price'
        fig.update_xaxes(row=1, col=1, tickcolor='white')
        fig.update_yaxes(title_text="Price", secondary_y=False, row=1, col=1, range=[min(df[price_col])-5, max(df[price_col])+5])
        fig.update_yaxes(title_text="Price", secondary_y=True, row=1, col=1, range=[min(df[price_col])-5, max(df[price_col])+5])
        fig.update_yaxes(title_text="Weights (in %)", secondary_y=False, row=2, col=1, range=[0,100+0.1])
        fig.update_yaxes(title_text="Weights (in %)", secondary_y=True, row=2, col=1, range=[0,100+0.1])
        fig.update_xaxes(title_text="Date", row=2, col=1)
        fig.update_layout(
            width=1500,
            height=1000,
            title=ticker+' (Weight Chart)' )

        return fig

    except Exception as e:
        print(e)



import warnings
warnings.filterwarnings('ignore')

import numpy as np
import pandas as pd
import yfinance as yf
import matplotlib.pyplot as plt
import dataframe_image as dfi

from .plots import return_plots, max_drawdown_plots, rolling_volatility_plots, rolling_max_drawdown_plots
from .plots import plotly_create_subplots, plotly_figure_show, general_plots, plot_returns_heatmap, annualised_periodic_returns, weight_graphs
from .finstats import financial_summary, detailed_summary
from .utils import detailed_summary


# heatmap color
import plotly.graph_objects as go
from plotly.subplots import make_subplots
import plotly.io as pio
pio.templates.default = 'simple_white'

import io

from reportlab.lib.pagesizes import letter
from reportlab.platypus import SimpleDocTemplate, Paragraph, Spacer, Image
from reportlab.lib.styles import getSampleStyleSheet
from reportlab.lib.units import inch

import numpy as np

import warnings
warnings.filterwarnings('ignore')


# data preprocessing and plot

class FinancialPlot():
    def __init__(self, detail_summary, strategy, optimization_path, asset_class, pr, period=12, frequency="M", Weights=False, display=True):
        # File names - path

        self.detail_summary = detail_summary
        self.strategy = strategy
        self.optimization_path = optimization_path
        self.asset_class = asset_class
        self.pr = pr
        self.period = period
        self.frequency = frequency
        self.weights_flag = Weights
        self.display = display



    def yearly_returns_data(self, detail_summary):
        self.yearly = detailed_summary(self.detail_summary, sheets=["Portfolio", "Baseline"], period=self.period, prev_yr=self.pr)
        self.yearly = self.yearly[:8]

        return self.yearly


    def detail_summary_data(self, detail_summary):
        self.portfolio = pd.read_excel(self.detail_summary, sheet_name= 'Portfolio', index_col = 0, engine='openpyxl')

        self.baseline = pd.read_excel(self.detail_summary, sheet_name = 'Baseline', engine='openpyxl')

        return self.portfolio, self.baseline

    # data preprocessing
    def statistic_data(self, portfolio, baseline):
        portfolio_stat = portfolio.copy()
        baseline_stat = baseline.copy()

        portfolio_stat.reset_index(inplace=True)

        baseline_stat['returns'] = baseline_stat['Adj Close'].pct_change().dropna()

        return portfolio_stat, baseline_stat

    def cum_ret_data(self, portfolio, baseline):

        baseline.set_index('date', inplace=True)
        cumm_return = pd.concat([portfolio[['invested']], baseline], axis = 1)

        cumm_return.rename(columns = {'invested': 'Portfolio', 'Adj Close' : 'Baseline'}, inplace = True)

        cumm_return.index = pd.to_datetime(cumm_return.index)

        # extra add for riskfolio
        cumm_return = cumm_return[['Portfolio', 'Baseline']]

        return cumm_return

    def calc_beta(self,df):
        np_array = df.values
        s = np_array[:, 0] # stock returns are column zero from numpy array
        m = np_array[:, 1] # market returns are column one from numpy array

        covariance = np.cov(s,m) # Calculate covariance between stock and market
        beta = covariance[0,1]/covariance[1,1]

        return beta

    def rolling_apply(self, df, period, func, min_periods=None):
        if min_periods is None:
            min_periods = period
        result = pd.Series(np.nan, index=df.index)

        for i in range(1, len(df)+1):
            sub_df = df.iloc[max(i-period, 0):i,:] #I edited here
            if len(sub_df) >= min_periods:
                idx = sub_df.index[-1]
                result[idx] = func(sub_df)
        return result

    def rolling_beta_data(self, portfolio, baseline, period_6m = 6, period_12m = 12):

        _portfolio = portfolio.iloc[1:-1]

        _baseline = baseline.iloc[:-1]

        _baseline.index = pd.to_datetime(_baseline.index)
        _baseline = _baseline.resample(self.frequency).last()
        _baseline['mkt_ret_1m'] = _baseline['Adj Close'].pct_change()
        _baseline.reset_index(inplace= True)


        rolling_beta = _portfolio.copy()
        rolling_beta.index = pd.to_datetime(rolling_beta.index)
        rolling_beta.reset_index(inplace=True)
        rolling_beta = pd.merge(rolling_beta, _baseline[['date','mkt_ret_1m']], on = 'date')
        rolling_beta = rolling_beta[['date', 'returns', 'mkt_ret_1m']]

        rolling_beta['6m rolling beta'] = self.rolling_apply(rolling_beta[['returns','mkt_ret_1m']], period_6m, self.calc_beta, min_periods = period_6m)
        rolling_beta['12m rolling beta'] = self.rolling_apply(rolling_beta[['returns','mkt_ret_1m']], period_12m, self.calc_beta, min_periods = period_12m)

        rolling_beta = rolling_beta[['date','6m rolling beta','12m rolling beta']]

        rolling_beta.set_index('date',inplace=True)

        return rolling_beta

    def max_drawdown_data(self, portfolio):
        max_drawdown = portfolio[['invested']]

        max_drawdown.rename(columns = {'invested': 'Portfolio'}, inplace = True)

        return max_drawdown

    def rolling_sharpe_data(self, portfolio):
        _portfolio = portfolio.iloc[1:-1]

        if self.frequency == "M":
            wind_6 = 6
            wind_12 = 12
        elif self.frequency == "W":
            wind_6 = 26
            wind_12 = 52
        elif self.frequency == "D":
            wind_6 = 126
            wind_12 = 252

        # test
        _portfolio['6 months rolling'] = _portfolio.returns.rolling(wind_6).apply(lambda x: (((1+x).prod()-1) / wind_6 * wind_12) / (x.std()*np.sqrt(wind_12)), raw = True)
        _portfolio['12 months rolling'] = _portfolio.returns.rolling(wind_12).apply(lambda x: (((1+x).prod()-1) / wind_12 * wind_12) / (x.std()*np.sqrt(wind_12)), raw = True)
        sharpe_ratio = pd.concat([_portfolio['6 months rolling'],_portfolio['12 months rolling']], axis = 1)

        return sharpe_ratio

    def heatmap_monthly_return_data(self, portfolio):

        # test
        monthly_ret_table = portfolio.copy()
        monthly_ret_table.reset_index(inplace=True)

        monthly_ret_table['date'] = pd.to_datetime(monthly_ret_table['date'])
        monthly_ret_table['year'] = monthly_ret_table['date'].dt.year
        monthly_ret_table['month'] = monthly_ret_table['date'].dt.month
        monthly_ret_table = monthly_ret_table[['year', 'month', 'returns']]
        monthly_ret_table['returns'] = monthly_ret_table['returns']*100
        monthly_ret_table['returns'] = monthly_ret_table['returns'].round(2)
        zmax = max(monthly_ret_table['returns'].max(), abs(monthly_ret_table['returns'].min()))


        monthly_ret_table = monthly_ret_table.set_index(['year', 'month']).unstack(level = -1)
        monthly_ret_table.columns = monthly_ret_table.columns.droplevel()
        monthly_ret_table.columns.name = None
        monthly_ret_table.rename({monthly_ret_table.columns[0]:'Jan', monthly_ret_table.columns[1]:'Feb',
                            monthly_ret_table.columns[2]: 'Mar', monthly_ret_table.columns[3]:'Apr',
                            monthly_ret_table.columns[4]:'May', monthly_ret_table.columns[5]:'Jun',
                            monthly_ret_table.columns[6]: 'Jul', monthly_ret_table.columns[7]:'Aug',
                            monthly_ret_table.columns[8]:'Sep', monthly_ret_table.columns[9]:'Oct',
                            monthly_ret_table.columns[10]:'Nov', monthly_ret_table.columns[11]:'Dec'},
                            axis = 1, inplace=True)
        monthly_ret_table = monthly_ret_table.transpose()

        return monthly_ret_table, zmax

    def annualized_data(self, yearly):

        yearly['Security Name'] = ['Portfolio', 'Portfolio', 'Portfolio', 'Portfolio', 'Baseline', 'Baseline', 'Baseline', 'Baseline']

        annualized = pd.concat([yearly.drop('Security Name', axis=1), yearly['Security Name']], axis=1)
        annualized = annualized.round(2)

        # calculation multiply 100 without sharpe ratio
        annualized = annualized.reset_index().set_index(['index', 'Security Name'])
        annualized.index.names = [None, 'Security Name']
        Y = pd.DataFrame(index=annualized.index, columns=annualized.columns)
        for i in annualized.index:
            for j in annualized.columns:
                if 'Sharpe Ratio' not in i:
                    Y.loc[i, j] = annualized.loc[i, j] * 100
                else:
                    Y.loc[i, j] = annualized.loc[i, j]
        Y = Y.reset_index().set_index('level_0')
        Y.index.name = None


        annualized = Y

        return annualized

    # Statistic and plot
    def statistic_plot(self, portfolio_stat, baseline_stat, strategy):

        data = financial_summary(portfolio_stat, baseline_stat, frequency=self.frequency, asset_class= self.asset_class, title = f'{self.strategy}')

        return data

    def cum_ret_plot(self, cumm_return):
        fig = return_plots(cumm_return, graph_width=800, graph_height=500)

        fig.update_xaxes()
        fig.update_layout(height=400, width=800, title_text=f"Cumulative Return {self.strategy}")

        return fig

    def max_drawdown_plot(self, max_drawdown,strategy):
        fig = max_drawdown_plots(max_drawdown, frequency= self.frequency, graph_width=800, graph_height=500, strategy = strategy)

        fig.update_layout(height=500, width=800, title_text=f"Maximum Drawdown {self.strategy}")

        return fig

    def rolling_beta_plot(self, rolling_beta):
        six_month_mean = rolling_beta['6m rolling beta'].mean()
        twelve_month_mean = rolling_beta['12m rolling beta'].mean()
        fig = general_plots(rolling_beta, frequency= self.frequency, graph_width=800, graph_height=500, strategy = self.strategy)


        fig.add_trace(go.Scatter(y = [six_month_mean, six_month_mean],
                                x = [min(rolling_beta.index), max(rolling_beta.index)],
                                mode = 'lines',
                                line = dict(width=2, dash='dash'),
                                marker_color='green',
                                name = f'Mean 6 months rolling beta')
                    )
        fig.add_trace(go.Scatter(y = [twelve_month_mean, twelve_month_mean],
                                 x = [min(rolling_beta.index), max(rolling_beta.index)],
                                 mode = 'lines',
                                 line = dict(width=2, dash='dash'),
                                 marker_color='red',
                                 name = f'Mean 12 months rolling beta')
                        )
        fig.update_layout(height=500, width=800, title_text=f"Rolling Portfolio Beta {self.strategy}")
        fig.update_xaxes(title_text="Date")
        fig.update_yaxes(title_text="Beta")

        return fig

    def rolling_sharpe_ratio_plot(self, sharpe_ratio):

        six_month_mean = sharpe_ratio['6 months rolling'].mean()
        twelve_month_mean = sharpe_ratio['12 months rolling'].mean()
        fig = general_plots(sharpe_ratio,frequency= self.frequency, graph_width=800, graph_height=500, strategy = self.strategy)


        fig.add_trace(go.Scatter(y = [six_month_mean, six_month_mean],
                                    x = [min(sharpe_ratio.index), max(sharpe_ratio.index)],
                                    mode = 'lines',
                                    line = dict(width=2, dash='dash'),
                                    marker_color='green',
                                    name = f'Mean 6 months rolling sharpe ratio')
                        )
        fig.add_trace(go.Scatter(y = [twelve_month_mean, twelve_month_mean],
                                    x = [min(sharpe_ratio.index), max(sharpe_ratio.index)],
                                    mode = 'lines',
                                    line = dict(width=2, dash='dash'),
                                     marker_color='red',
                                    name = f'Mean 12 months rolling sharpe ratio')
                        )
        fig.update_layout(height=500, width=800, title_text=f"Rolling Sharpe Ratio {self.strategy}")
        fig.update_xaxes(title_text="Date")
        fig.update_yaxes(title_text="Sharpe Ratio")

        return fig

    def heatmap_monthly_return_plot(self, monthly_ret_table, zmax):

        fig = plot_returns_heatmap(monthly_ret_table,frequency = self.frequency, graph_width=800, graph_height=500, strategy = self.strategy, z = zmax)

        fig.update_xaxes(title_text="Year")
        fig.update_yaxes(title_text="Month")

        fig.update_layout(coloraxis = {'colorscale':'RdYlGn'})
        fig.update_xaxes(side="bottom")
        fig.update_coloraxes(colorbar_title_text = 'Return(%)')
        fig.update_layout(
              title=f'Monthly Return {self.strategy}',
              plot_bgcolor='white'
            )

        return fig


    def annualized_plot(self, ann, strategy):

        fig = annualised_periodic_returns(ann, securities = ['Portfolio', 'Baseline'], frequency= self.frequency, graph_width=1200, graph_height=100, strategy = self.strategy)

        fig.update_layout(height=500, width=900, title_text=f"Yearwise Annual Return {self.strategy}")

        return fig

    def weight_plot(self, tic, strategy):
        try:
            os.mkdir('./Price Data')
        except:
            pass

        tickers = list(pd.read_excel(self.detail_summary, sheet_name="Weights", index_col=0).columns)

        for tick in tickers:
            data = yf.download(tick, start="2007-01-01", progress=False)
            data['TIC']= tick
            data.to_csv(f"Price Data/{tick}.csv")

        wfig = weight_graphs(weights_file = self.detail_summary,
                           ticker=tic,
                           threshold = 0.01,
                           path='./Price Data',
                           price_col='Adj Close',
                           frequency=self.frequency)

        wfig.update_layout(width=900,
                           height=500,
                           title=f'Weight and Signal Chart ('+tic+')')

        return wfig


    def main_(self):

        # read
        image_path = []
        portfolio, baseline = self.detail_summary_data(self.detail_summary)
        yearly_ = self.yearly_returns_data(self.detail_summary)
        yearly = self.yearly_returns_data(yearly_) #####

        # data
        portfolio_stat, baseline_stat = self.statistic_data(portfolio, baseline)
        cumm_return = self.cum_ret_data(portfolio, baseline)
        rolling_beta = self.rolling_beta_data(portfolio, baseline)
        sharpe_ratio = self.rolling_sharpe_data(portfolio)
        max_drawdown = self.max_drawdown_data(portfolio)
        monthly_ret_table, zmax = self.heatmap_monthly_return_data(portfolio)
        ann = self.annualized_data(yearly)

        # plot

        data = self.statistic_plot(portfolio_stat, baseline_stat, self.strategy)
        dfi.export(data, f'{self.optimization_path}/test.png')
        image_path.append(f'./{self.optimization_path}/test.png')

        image_path.append(f'./{self.optimization_path}/blank.png')


        from .finstats import financial_summary, detailed_summary
        dfi.export(detailed_summary(self.detail_summary, sheets = ['Portfolio', 'Baseline'], prev_yr=self.pr, period = self.period, title=f'YearWise Summary | {self.strategy}'), f'{self.optimization_path}/test_yearly.png')
        image_path.append(f'./{self.optimization_path}/test_yearly.png')

        cum_ret_fig = self.cum_ret_plot(cumm_return)
        cum_ret_fig.write_image(f'{self.optimization_path}/cum_ret_plot.png')
        image_path.append(f'./{self.optimization_path}/cum_ret_plot.png')


        max_drawdown_fig = self.max_drawdown_plot(max_drawdown, self.strategy)
        max_drawdown_fig.write_image(f'{self.optimization_path}/max_drawdown_plot.png')
        image_path.append(f'./{self.optimization_path}/max_drawdown_plot.png')


        rolling_beta_fig = self.rolling_beta_plot(rolling_beta)
        rolling_beta_fig.write_image(f'{self.optimization_path}/rolling_beta_plot.png')
        image_path.append(f'./{self.optimization_path}/rolling_beta_plot.png')


        rolling_sharpe_fig = self.rolling_sharpe_ratio_plot(sharpe_ratio)
        rolling_sharpe_fig.write_image(f'{self.optimization_path}/rolling_sharpe_plot.png')
        image_path.append(f'./{self.optimization_path}/rolling_sharpe_plot.png')


        heatmap_monthly_return_fig = self.heatmap_monthly_return_plot(monthly_ret_table, zmax)
        heatmap_monthly_return_fig.write_image(f'{self.optimization_path}/heatmap_monthly_plot.png')
        image_path.append(f'./{self.optimization_path}/heatmap_monthly_plot.png')


        ann_fig = self.annualized_plot(ann, self.strategy)
        ann_fig.write_image(f'{self.optimization_path}/ann_plot_fig.png')
        image_path.append(f'./{self.optimization_path}/ann_plot_fig.png')

        yearwise_summary = detailed_summary(self.detail_summary, sheets = ['Portfolio', 'Baseline'], prev_yr=self.pr, title=f'{self.strategy}')


        if self.display:
            display(detailed_summary(self.detail_summary, sheets = ['Portfolio', 'Baseline'], prev_yr=self.pr, period = self.period, title=f'YearWise Summary | {self.strategy}'))
            display(data)

            cum_ret_fig.show()
            max_drawdown_fig.show()
            rolling_beta_fig.show()
            rolling_sharpe_fig.show()
            heatmap_monthly_return_fig.show()
            ann_fig.show()




        tickers = list(pd.read_excel(self.detail_summary, sheet_name="Weights", index_col=0).columns)

        if self.weights_flag:
            for tic in tickers:
                wfig = self.weight_plot(tic, self.strategy)
                wfig.write_image(f'{self.optimization_path}/weight_{tic}_fig.png')
                image_path.append(f'./{self.optimization_path}/weight_{tic}_fig.png')

        return image_path



import numpy as np
import pandas as pd
import pymannkendall as mk
import statsmodels.api as sm
import matplotlib.pyplot as plt
from scipy.stats import kurtosis


plt.rcParams['axes.facecolor'] = 'lightblue'


def CAGR(df:pd.DataFrame, t:float, col_name:str="invested") -> float:
    '''
    Calculates the Cumulative Annual Growth Return

    Parameters
    ----------
    df : pd.DataFrame, default - None
        Expects a dataframe which has the invested amounts
    t : str, default - float
        Time period in years
    col_name : str, default - "invested"
        Name of the columns

    Return
    ------
    float : A float value (CAGR)
    '''
    assert (str(type(df)) == "<class 'pandas.core.frame.DataFrame'>"), "Not a dataframe"
    assert (str(type(t)) == "<class 'numpy.float64'>"), "Not a float values"
    assert (str(type(col_name)) == "<class 'str'>"), "Not a string"

    try:
        return (((df[col_name].iloc[-1] / df[col_name].iloc[0]) ** (1/t)) - 1) * 100
    except Exception as e:
        print(e)





def information_ratio(returns:pd.Series, benchmark_returns:pd.Series, period:int) -> float:
    '''
    Calculates the Information Ratio

    Parameters
    ----------
    returns : pd.Series, default - None
        Expects a series containing the return series
    benchmark_returns : pd.Series, default - None
        Excepts a series containing the market returns
    period : int, default - None
        The period (monthly or daily etc.)

    Return
    ------
    float : Information Ratio
    '''
    assert (str(type(returns)) == "<class 'pandas.core.series.Series'>"), "Not a Series"
    assert (str(type(period)) == "<class 'int'>"), "Not an integer value"
    assert (str(type(benchmark_returns)) == "<class 'pandas.core.series.Series'>"), "Not a Series"

    try:
        return_difference = returns - benchmark_returns
        volatility = return_difference.std() * np.sqrt(period)
        information_ratio = return_difference.mean() / volatility

        return information_ratio
    except Exception as e:
        print(e)






def financial_summary(df_rets:pd.DataFrame, benchmark_rets:pd.DataFrame, frequency:str='D', asset_class:str='None', risk_free_rate:float=0, col_name_cagr:str="invested", title:str='', date_col:str='date') -> pd.DataFrame:
    '''
    Must supply a dataframe with date, retruns as columns
    Note - Don't supply daily returns as % . Keep the date column at the beginning.

    Example:
    date           returns    turnover
    2018-02-09     0.25       0.25
    2018-02-10     0.29       0.00

    Parameters
    ----------
    frequency : Daily (D), Monthly(M) or Weekly(W), default - 'D'.
            Describes the frequency of data provided.
    df_rets : pd.DataFrame, default - None
            All Return Information, Turnover etc.
    asset_class : str, default - ''
            Name of the asset class
    risk_free_rate : float, default 0
            risk free rate of return
    col_name_cagr : str, default - "invested"
            The name of the column for CAGR Calculation (column should be present in df_rets)
    benchmark_rets : pd.DataFrame, default - None
            Market returns for the given period
    date_col : str, default - 'date'
            Name of the date column

    Return
    ------
    pd.DataFrame : Complete Summary of any strategy
    '''
    assert (str(type(df_rets)) == "<class 'pandas.core.frame.DataFrame'>"), "Not a dataframe"
    assert (str(type(benchmark_rets)) == "<class 'pandas.core.frame.DataFrame'>"), "Not a dataframe"
    assert (str(type(col_name_cagr)) == "<class 'str'>"), "Not a string"
    assert (str(type(asset_class)) == "<class 'str'>"), "Not a string"
    assert (frequency in ['D', 'M', 'W', 'Q', 'Y', '6M']), "Should be one of the values of 'D', 'M', 'W', 'Q', 'Y', '6M'"
    assert (str(type(title)) == "<class 'str'>"), "Not a string"

    try:
        mapper = {'D': 'Day', 'M': 'Month', 'W': 'Week', 'Q': 'Quarter', 'Y': 'Year', '6M': 'Half Yearly'}
        mapper_time = {'D': 252, 'M': 12, 'W': 52, 'Q': 4, 'Y': 1, '6M': 2}

        # If turnover is provided we skip, else we take it as 0.
        try:
            Y = df_rets['turnover']
        except:
            df_rets['turnover'] = 0

        df_rets['c_ret'] = (1 + df_rets['returns']).cumprod() - 1
        periods = (pd.to_datetime(df_rets.loc[df_rets.iloc[len(df_rets)-1].name, date_col]) - pd.to_datetime(df_rets.loc[df_rets.iloc[0].name, date_col])) // np.timedelta64(1, frequency)
        volatility = np.std(df_rets['returns']) * np.sqrt(mapper_time[frequency])
        returns = ((df_rets['c_ret'].values[-1])/(periods)) * mapper_time[frequency]
        sharpe = (returns - risk_free_rate) / volatility
        cagr = CAGR(df_rets, t=np.around((periods/mapper_time[frequency]), 2), col_name=col_name_cagr)
        info_ratio = information_ratio(df_rets['returns'], benchmark_rets['returns'], mapper_time[frequency])
        turnover = df_rets['turnover'].mean()
        market_class = asset_class
        drawdowns = drawdown(df_rets['returns'])['Drawdown'].min()
        sortino = returns / (np.std(df_rets[df_rets['returns'] < 0]['returns']) * np.sqrt(mapper_time[frequency]))
        kurtosiss = kurtosis(df_rets['returns'], fisher=False)

        # Data Arrangement
        meta_data = pd.DataFrame(
            data = [
                df_rets.loc[df_rets.iloc[0].name, date_col],
                df_rets.loc[df_rets.iloc[len(df_rets)-1].name, date_col],
                periods,
                asset_class
            ],

            index = ['Start Date', 'End Date', f'Time Period (in {mapper[frequency]})', 'Strategy'], columns=['Meta Data']
        )

        summary_data = pd.DataFrame(
            data = [
                f'{np.around(returns*100, 2)}%',
                f'{np.around(volatility*100, 2)}%',
                f'{np.around(cagr, 2)}%',
                f'{np.around(sortino, 2)}'
            ],

            index = ['Annual Return', 'Annual Volatility', f'CAGR', 'Sortino Ratio'], columns=['Summary']
        )

        other_stats = pd.DataFrame(
            data = [
                f'{np.around(sharpe, 2)}',
                f'{np.around(kurtosiss, 2)}',
                f'{np.around(info_ratio, 2)}',
                f'{np.around(turnover*100, 2)}%'
            ],

            index = ['Sharpe Ratio', 'Kurtosis', f'Information Ratio', 'Turnover'], columns=['Statistics']
        )

        meta_data = meta_data.reset_index()
        summary_data = summary_data.reset_index()
        other_stats = other_stats.reset_index()

        meta_data.columns = ['', 'Meta Data']
        other_stats.columns = ['', 'Statistics']
        summary_data.columns = ['', 'Summary']

        df = pd.concat([meta_data, summary_data, other_stats], axis=1)
        df = df.style.set_caption(f'<b>{title}</b>')

        return df

    except Exception as e:
        print(e)





def drawdown(return_series: pd.Series) -> pd.DataFrame:
    """Takes a time series of asset returns.
       returns a DataFrame with columns for
       the wealth index,
       the previous peaks, and
       the percentage drawdown

       Parameter
       --------
       returns_series : pd.Series, default - None

       Return
       ------
       pd.DataFrame : Drawdown
    """
    assert (str(type(return_series)) == "<class 'pandas.core.series.Series'>"), "Not a Series"

    try:
        return_series = return_series.apply(lambda x: np.log(1+x))
        wealth_index = 1000 * (1+return_series).cumprod()
        previous_peaks = wealth_index.cummax()
        drawdowns = (wealth_index - previous_peaks)/previous_peaks
        return pd.DataFrame({"Wealth": wealth_index,
                            "Previous Peak": previous_peaks,
                            "Drawdown": drawdowns})

    except Exception as e:
        print(e)





def detailed_summary(excelSheet:str, sheets:list, col_name="invested", period=12, prev_yr='2017', title:str='') -> pd.DataFrame:
    '''
    Provides an year-wise summary of portfolio, baseline and the individual securities.

    Parameters
    ----------
    excelSheet: str, default - ''
        Name of the ExcelSheet with the details.
    sheets: list
        A list with sheet Names - ["Portfolio", "Baseline", "Prices"]
    period: int, default - 12
        Data details, whether monthly, daily or quarterly data. Default is 12 i.e. monthly, change it to 4 if quarterly or 252 if daily.
    prev_yr: str, default - '2017'
        Last year preceeding the first date. For Ex: if first date is "2018-03-01", prev_yr = 2017 (Since the first date entry is on 2018)
    col_name: str, default - 'invested'
        Name of the column on which statistics are calculated for the Portfolio Sheet.

    Note: "Provide the entire path to the excelsheet if it is inside a folder."
          "All sheets must have a column named "date" for dates."
          "Adj Close" - In Baseline Sheet.(In Baseline sheet)"

    Return
    ------
    pd.DataFrame: A detailed summary (yearwise)
    '''
    assert (str(type(period)) == "<class 'int'>"), "Not an integer value"
    assert (str(type(prev_yr)) == "<class 'str'>"), "Not a string"
    assert (str(type(excelSheet)) == "<class 'str'>"), "Not a string"
    assert (str(type(col_name)) == "<class 'str'>"), "Not a string"
    assert (str(type(sheets)) == "<class 'list'>"), "Not a list object"
    assert (str(type(title)) == "<class 'str'>"), "Not a string"


    try:
        def individual_etfs(df):
            Y = list(df.columns[1:len(df.columns)-1])
            dfx = pd.DataFrame()

            # for all etfs in the portfolio
            for etf in Y:
                tf = pd.DataFrame()
                prev_date = prev_yr
                for date in df['year'].unique():
                    y = df[(df_port.date>=prev_date+'-12-31') & (df_port.date<=date+'-12-31')].loc[:, [etf]]
                    y.reset_index(drop=True, inplace=True)

                    # Date update
                    prev_date = date

                    # Log Returns
                    lgrets = np.diff(np.log(y[etf]))
                    lgrets = np.insert(lgrets, 0, np.nan)
                    y['log_returns'] = lgrets
                    y['log_returns'].fillna(0, inplace=True)

                    # Volatility
                    vol = np.std(y.loc[1:, 'log_returns'], ddof=1) * np.sqrt(period)

                    # Annual Return
                    ret = (y.loc[len(y)-1, etf] / y.loc[0, etf]) - 1

                    # Max Drawdown
                    inv = y[etf]
                    z = pd.Series(index=range(len(inv)))
                    z.iloc[0] = inv.iloc[0]

                    for i in range(1, len(inv)):
                        z.iloc[i] = max(inv[i], z[i-1])

                    maxdrwdn = (inv - z).min()/z[0]

                    # Sharpe ratio
                    sharpe = ret / vol

                    new_df = pd.DataFrame(data=[f'{np.around(ret*100, 2)}%', f'{np.around(vol*100, 2)}%', f'{np.around(sharpe, 2)}', f'{np.around(maxdrwdn*100, 2)}%'], columns=[date], index=['Return', 'Volatility', 'Sharpe Ratio', 'Max Drawdown'])
                    tf = pd.concat([tf, new_df], axis=1)

                tf.columns.name = etf
                tf['Security Name'] = etf
                dfx = pd.concat([dfx, tf])

            return dfx


        # Read the entire excel sheet
        portfolio_returns, baseline_returns = [], []

        result = pd.ExcelFile(excelSheet)
        df = pd.DataFrame()
        prev_date = prev_yr

        for ticker in sheets:
            df_port = pd.read_excel(result, ticker)
            df_port['year'] = df_port['date'].apply(lambda x: x[:4])
            tf = pd.DataFrame()

            if ticker == 'Portfolio':
                for date in df_port['year'].unique():
                    y = df_port[(df_port.date>=prev_date+'-12-31') & (df_port.date<=date+'-12-31')]
                    y.reset_index(drop=True, inplace=True)

                    # Date update
                    prev_date = date

                    # Log Returns
                    lgrets = np.diff(np.log(y[col_name]))
                    lgrets = np.insert(lgrets, 0, np.nan)
                    y['log_returns'] = lgrets
                    y['log_returns'].fillna(0, inplace=True)

                    # Volatility
                    vol = np.std(y.loc[1:, 'log_returns'], ddof=1) * np.sqrt(period)

                    # Annual Return
                    ret = (y.loc[len(y)-1, col_name] / y.loc[0, col_name]) - 1

                    # For alpha calculation
                    portfolio_returns.append(ret)

                    # Max Drawdown
                    inv = y[col_name]
                    z = pd.Series(index=range(len(inv)))
                    z.iloc[0] = inv.iloc[0]

                    for i in range(1, len(inv)):
                        z.iloc[i] = max(inv[i], z[i-1])

                    maxdrwdn = (inv - z).min()/z[0]

                    # Sharpe ratio
                    sharpe = ret / vol

                    new_df = pd.DataFrame(data=[f'{np.around(ret*100, 2)}%', f'{np.around(vol*100, 2)}%', f'{np.around(sharpe, 2)}', f'{np.around(maxdrwdn*100, 2)}%'], columns=[date], index=['Return', 'Volatility', 'Sharpe Ratio', 'Max Drawdown'])
                    tf = pd.concat([tf, new_df], axis=1)

                tf.columns.name = ticker
                tf['Security Name'] = ticker
                df = pd.concat([df, tf])

            elif ticker == 'Baseline':
                prev_date = prev_yr

                for date in df_port['year'].unique():
                    y = df_port[(df_port.date>=prev_date+'-12-31') & (df_port.date<=date+'-12-31')]
                    y.reset_index(drop=True, inplace=True)

                    # Date update
                    prev_date = date

                    # Log Returns
                    lgrets = np.diff(np.log(y['Adj Close']))
                    lgrets = np.insert(lgrets, 0, np.nan)
                    y['log_returns'] = lgrets
                    y['log_returns'].fillna(0, inplace=True)

                    # Volatility
                    vol = np.std(y.loc[1:, 'log_returns'], ddof=1) * np.sqrt(period)

                    # Annual Return
                    ret = (y.loc[len(y)-1, 'Adj Close'] / y.loc[0, 'Adj Close']) - 1

                    # For alpha calculation
                    baseline_returns.append(ret)

                    # Max Drawdown
                    inv = y['Adj Close']
                    z = pd.Series(index=range(len(inv)))
                    z.iloc[0] = inv.iloc[0]

                    for i in range(1, len(inv)):
                        z.iloc[i] = max(inv[i], z[i-1])

                    maxdrwdn = (inv - z).min()/z[0]

                    # Sharpe ratio
                    sharpe = ret / vol

                    new_df = pd.DataFrame(data=[f'{np.around(ret*100, 2)}%', f'{np.around(vol*100, 2)}%', f'{np.around(sharpe, 2)}', f'{np.around(maxdrwdn*100, 2)}%'], columns=[date], index=['Return', 'Volatility', 'Sharpe Ratio', 'Max Drawdown'])
                    tf = pd.concat([tf, new_df], axis=1)

                tf.columns.name = ticker
                tf['Security Name'] = ticker
                df = pd.concat([df, tf])

            else:
                tf = individual_etfs(df_port)
                df = pd.concat([df, tf])

        # Set Colors
        def _color_negative_red(val):
            try:
                color = 'red' if float(val.split('%')[-2]) < 0 else 'green'
                return 'color: %s' % color
            except:
                color = 'red' if float(val) < 0 else 'green'
                return 'color: %s' % color


        if len(sheets) > 1 or sheets[0] == 'Prices':
            _return = df.loc['Return'].reset_index().set_index(['index', 'Security Name'])
            _volatility = df.loc['Volatility'].reset_index().set_index(['index', 'Security Name'])
            _sharpe_ratio = df.loc['Sharpe Ratio'].reset_index().set_index(['index', 'Security Name'])
            _max_drawdown = df.loc['Max Drawdown'].reset_index().set_index(['index', 'Security Name'])

            if (len(portfolio_returns) == len(baseline_returns)) and (len(portfolio_returns) > 0):
                alphas = pd.Series(np.array(portfolio_returns) - np.array(baseline_returns)).apply(lambda x: f'{np.around(x * 100, 2)}%')
                _return.loc[('Return', 'Alpha'), :] = alphas.values

            final_df = pd.concat([_return, _volatility, _sharpe_ratio, _max_drawdown], axis = 0)
            final_df.index.names = ['Metric', 'Security Name']

        elif len(sheets) == 1 and (sheets[0] != 'Prices'):
            final_df = df.copy()
            final_df.drop('Security Name', axis=1, inplace=True)

        final_df = final_df.style.applymap(_color_negative_red).set_caption(f'<b>{title}</b>')

        return final_df

    except Exception as e:
        print(e)



# Importing required libraries

import pandas as pd
import numpy as np
import talib as tb



# Calculate SMA for given periods

def SMA(close:pd.Series, short_period:int, long_period:int) -> pd.DataFrame:

    """
    Here we calculate the SMA value for given long and short time period.
    close:pd.Series = Price of the stock
    short_period:int = Short time period
    long_period:int = Long time period

    """
    # shortSMA is the sma of short_period
    shortSMA = tb.SMA(close,short_period)
    # longSMA is the sma of the long_period
    longSMA = tb.SMA(close,long_period)
    # Generating SMA signal : 1 when the difference between shortSMA and longSMA is positive else it is -1.
    smaSignal = [1 if (short - long) > 0 else (-1 if(short - long) < 0 else 0) for long, short in zip(longSMA, shortSMA)]


    # setting column names
    cols = ['Adj Close','Signal', 'shortSMA', 'longSMA']
    df  =  pd.DataFrame(columns=cols)
    # Assigning Close price to df['Close']
    df['Adj Close'] = close
    # Assigning SMA to df['Signal']
    df['Signal'] = smaSignal
    # Assigning Short SMA to df['shortSMA']
    df['shortSMA'] = shortSMA
    # Assigning Long SMA to df['longSMA']
    df['longSMA'] = longSMA
    df = df.fillna(0)

    return df



# Calculate EMA for given periods

def EMA(close:pd.Series, short_period:int, long_period:int) -> pd.DataFrame:

    """
    Here we calculate the EMA value for given long and short time period.
    close:pd.Series = Price of the stock
    short_period:int = Short time period
    long_period:int = Long time period

    """
    # shortEMA is the sma of short_period
    shortEMA = tb.EMA(close,short_period)
    # longEMA is the sma of the long_period
    longEMA = tb.EMA(close,long_period)
    # Generating EMA signal : 1 when the difference between shortEMA and longEMA is positive else it is -1.
    emaSignal = [1 if (short - long) > 0 else (-1 if (short - long) < 0 else 0) for long, short in zip(longEMA, shortEMA)]

    # setting column names
    cols = ['Adj Close','Signal', 'shortEMA', 'longEMA']
    df  =  pd.DataFrame(columns=cols)
    # Assigning Close price to df['Close']
    df['Adj Close'] = close
    # Assigning EMA to df['Signal']
    df['Signal'] = emaSignal
    # Assigning Short EMA to df['shortEMA']
    df['shortEMA'] = shortEMA
    # Assigning Long EMA to df['longEMA']
    df['longEMA'] = longEMA
    df = df.fillna(0)

    return df



# Calculate RSI for given periods

def RSI(close:pd.Series, short_period:int, long_period:int) -> pd.DataFrame:

    """
    Here we calculate the RSI value for given long and short time period.
    close:pd.Series = Price of the stock
    short_period:int = Short time period
    long_period:int = Long time period

    """
    # shortRSI is the rsi of short_period
    shortRSI = tb.RSI(close,timeperiod=short_period)
    # longRSI is the rsi of the long_period
    longRSI = tb.RSI(close,timeperiod=long_period)
    # Generating RSI signal : 1 when the difference between shortRSI and longRSI is positive else it is -1.
    rsiSignal = [1 if (short - long) > 0 else (-1 if (short - long) < 0 else 0) for long, short in zip(longRSI, shortRSI)]

    # setting column names
    cols = ['Adj Close','Signal', 'shortRSI', 'longRSI']
    df  =  pd.DataFrame(columns=cols)
    # Assigning Close price to df['Close']
    df['Adj Close'] = close
    # Assigning RSI to df['Signal']
    df['Signal'] = rsiSignal
    # Assigning Short RSI to df['shortRSI']
    df['shortRSI'] = shortRSI
    # Assigning Long RSI to df['longRSI']
    df['longRSI'] = longRSI
    df = df.fillna(0)

    return df



# Calculate ROC for given periods

def ROC(close:pd.Series, short_period:int, long_period:int) -> pd.DataFrame:

    """
    Here we calculate the ROC value for given long and short time period.
    close:pd.Series = Price of the stock
    short_period:int = Short time period
    long_period:int = Long time period

    """
    # shortROC is the roc of short_period
    shortROC = tb.ROC(close,timeperiod=short_period)
    # longROC is the roc of the long_period
    longROC = tb.ROC(close,timeperiod=long_period)
    # Generating ROC signal : 1 when the difference between shortROC and longROC is positive else it is -1.
    rocSignal = [1 if (short - long) > 0 else (-1 if (short - long) < 0 else 0 ) for long, short in zip(longROC, shortROC)]

    # setting column names
    cols = ['Adj Close','Signal', 'shortROC', 'longROC']
    df  =  pd.DataFrame(columns=cols)
    # Assigning Close price to df['Close']
    df['Adj Close'] = close
    # Assigning ROC to df['Signal']
    df['Signal'] = rocSignal
    # Assigning Short ROC to df['shortROC']
    df['shortROC'] = shortROC
    # Assigning Long ROC to df['longROC']
    df['longROC'] = longROC
    df = df.fillna(0)

    return df



# Profit and Loss calculator

def PnL(data:pd.DataFrame, price_col:str="Adj Close") -> float:

    """
    Here we calculate the total profit and loss, given the Price and the Signal.
    data: pd.DataFrame = Price and Signal columns dataframe

    """
    # Calculating the P/L by multiplying the price difference with the signal and summing up to get the whole P/L.
    pnl = (data[price_col].diff(1) * data['Signal'].shift(1)).drop(0).sum()

    return pnl



# Technical Indicator Optimizer

def optimizeTI(price:pd.Series, min_long:int, max_long:int, min_short:int, max_short:int, TI:str ) -> tuple:

    """
    Here we optimize the long and short time period given, the price, long range - (min ,max) and short range - (min ,max).

    price: pd.Series = Price of the stock
    min_long:int = minimum value of the long time period
    max_long:int = maximum value of the long time period
    min_short:int = minimum value of the short time period
    max_short:int = maximum value of the short time period
    TI:str = Technical indicator that you want to optimize

    returns
    -------
    long, short: Long and Short Time Periods

    """
    # opt_short = optimized short time period
    opt_short = -1
    # opt_long = optimized long time period
    opt_long = -1
    # initializing maximum P/L as 0
    max_pnl = 0

    # iterating over the long range.
    for i in range(min_long, max_long):

        # iterating over the short range.
        for j in range(min_short, max_short):
            # Checking short value
            if j > i :
                # Skipping if short value is more than long value
                continue
            else:
                # Calculate TI for given periods
                if TI == 'SMA':
                    df = SMA(price, j, i)
                elif TI == 'EMA':
                    df = EMA(price, j, i)
                elif TI == 'RSI':
                    df = RSI(price, j, i)
                elif TI == 'ROC':
                    df = ROC(price, j, i)

                # Calculating P/L based on calculated TI
                pnl = PnL(df)

                if (opt_short == -1) or (opt_long  == -1):
                    # Updating Optimized short and long periods
                    max_pnl = pnl
                    opt_short = j
                    opt_long = i

                if max_pnl < pnl:
                    # Updating Optimized short and long periods
                    max_pnl = pnl
                    opt_short = j
                    opt_long = i

    return opt_long, opt_short



import pandas as pd
import scipy.stats
from scipy.stats import norm
import numpy as np
from scipy.optimize import minimize
import math


def skewness(r):
    """
    Alternative to scipy.stats.skew()
    Computes the skewness of the supplied Series or DataFrame
    Returns a float or a Series
    """
    demeaned_r = r - r.mean()
    # use the population standard deviation, so set dof=0
    sigma_r = r.std(ddof=0)
    exp = (demeaned_r**3).mean()
    return exp/sigma_r**3


def kurtosis(r):
    """
    Alternative to scipy.stats.kurtosis()
    Computes the kurtosis of the supplied Series or DataFrame
    Returns a float or a Series
    """
    demeaned_r = r - r.mean()
    # use the population standard deviation, so set dof=0
    sigma_r = r.std(ddof=0)
    exp = (demeaned_r**4).mean()
    return exp/sigma_r**4


def is_normal(r, level=0.01):
    """
    Applies the Jarque-Bera test to determine if a Series is normal or not
    Test is applied at the 1% level by default
    Returns True if the hypothesis of normality is accepted, False otherwise
    """
    if isinstance(r, pd.DataFrame):
        return r.aggregate(is_normal)
    else:
        statistic, p_value = scipy.stats.jarque_bera(r)
        return p_value > level

def semideviation(r):
    """
    Returns the semideviation aka negative semideviation of r
    r must be a Series or a DataFrame, else raises a TypeError
    """
    if isinstance(r, pd.Series):
        is_negative = r < 0
        return r[is_negative].std(ddof=0)
    elif isinstance(r, pd.DataFrame):
        return r.aggregate(semideviation)
    else:
        raise TypeError("Expected r to be a Series or DataFrame")


def var_historic(r, level=5):
    """
    Returns the historic Value at Risk at a specified level
    i.e. returns the number such that "level" percent of the returns
    fall below that number, and the (100-level) percent are above
    """
    if isinstance(r, pd.DataFrame):
        return r.aggregate(var_historic, level=level)
    elif isinstance(r, pd.Series):
        return -np.percentile(r, level)
    else:
        raise TypeError("Expected r to be a Series or DataFrame")


def cvar_historic(r, level=5):
    """
    Computes the Conditional VaR of Series or DataFrame
    """
    if isinstance(r, pd.Series):
        is_beyond = r <= -var_historic(r, level=level)
        return -r[is_beyond].mean()
    elif isinstance(r, pd.DataFrame):
        return r.aggregate(cvar_historic, level=level)
    else:
        raise TypeError("Expected r to be a Series or DataFrame")


def var_gaussian(r, level=5, modified=False):
    """
    Returns the Parametric Gauusian VaR of a Series or DataFrame
    If "modified" is True, then the modified VaR is returned,
    using the Cornish-Fisher modification
    """
    # compute the Z score assuming it was Gaussian
    z = norm.ppf(level/100)
    if modified:
        # modify the Z score based on observed skewness and kurtosis
        s = skewness(r)
        k = kurtosis(r)
        z = (z +
                (z**2 - 1)*s/6 +
                (z**3 -3*z)*(k-3)/24 -
                (2*z**3 - 5*z)*(s**2)/36
            )
    return -(r.mean() + z*r.std(ddof=0))


def portfolio_return(weights, returns):
    """
    Computes the return on a portfolio from constituent returns and weights
    weights are a numpy array or Nx1 matrix and returns are a numpy array or Nx1 matrix
    """
    return weights.T @ returns


def portfolio_vol(weights, covmat):
    """
    Computes the vol of a portfolio from a covariance matrix and constituent weights
    weights are a numpy array or N x 1 maxtrix and covmat is an N x N matrix
    """
    return (weights.T @ covmat @ weights)**0.5


def minimize_vol(target_return, er, cov):
    """
    Returns the optimal weights that achieve the target return
    given a set of expected returns and a covariance matrix
    """
    n = er.shape[0]
    init_guess = np.repeat(1/n, n)
    bounds = ((0.0, 1.0),) * n # an N-tuple of 2-tuples!
    # construct the constraints
    weights_sum_to_1 = {'type': 'eq',
                        'fun': lambda weights: np.sum(weights) - 1
    }
    return_is_target = {'type': 'eq',
                        'args': (er,),
                        'fun': lambda weights, er: target_return - portfolio_return(weights,er)
    }
    weights = minimize(portfolio_vol, init_guess,
                       args=(cov,), method='SLSQP',
                       options={'disp': False},
                       constraints=(weights_sum_to_1,return_is_target),
                       bounds=bounds)
    return weights.x



def msr(riskfree_rate, er, cov):
    """
    Returns the weights of the portfolio that gives you the maximum sharpe ratio
    given the riskfree rate and expected returns and a covariance matrix
    """
    n = er.shape[0]
    init_guess = np.repeat(1/n, n)
    bounds = ((0.0, 1.0),) * n # an N-tuple of 2-tuples!
    # construct the constraints
    weights_sum_to_1 = {'type': 'eq',
                        'fun': lambda weights: np.sum(weights) - 1
    }
    def neg_sharpe(weights, riskfree_rate, er, cov):
        """
        Returns the negative of the sharpe ratio
        of the given portfolio
        """
        r = portfolio_return(weights, er)
        vol = portfolio_vol(weights, cov)
        return -(r - riskfree_rate)/vol
    
    weights = minimize(neg_sharpe, init_guess,
                       args=(riskfree_rate, er, cov), method='SLSQP',
                       options={'disp': False},
                       constraints=(weights_sum_to_1,),
                       bounds=bounds)
    return weights.x


def gmv(cov):
    """
    Returns the weights of the Global Minimum Volatility portfolio
    given a covariance matrix
    """
    n = cov.shape[0]
    return msr(0, np.repeat(1, n), cov)


def optimal_weights(n_points, er, cov):
    """
    Returns a list of weights that represent a grid of n_points on the efficient frontier
    """
    target_rs = np.linspace(er.min(), er.max(), n_points)
    weights = [minimize_vol(target_return, er, cov) for target_return in target_rs]
    return weights


def pv(flows, r):
    """
    Compute the present value of a sequence of cash flows given by the time (as an index) and amounts
    r can be a scalar, or a Series or DataFrame with the number of rows matching the num of rows in flows
    """
    dates = flows.index
    discounts = discount(dates, r)
    return discounts.multiply(flows, axis='rows').sum()

def funding_ratio(assets, liabilities, r):
    """
    Computes the funding ratio of a series of liabilities, based on an interest rate and current value of assets
    """
    return pv(assets, r)/pv(liabilities, r)

def inst_to_ann(r):
    """
    Convert an instantaneous interest rate to an annual interest rate
    """
    return np.expm1(r)

def ann_to_inst(r):
    """
    Convert an instantaneous interest rate to an annual interest rate
    """
    return np.log1p(r)


def discount(t, r):
    """
    Compute the price of a pure discount bond that pays a dollar at time period t
    and r is the per-period interest rate
    returns a |t| x |r| Series or DataFrame
    r can be a float, Series or DataFrame
    returns a DataFrame indexed by t
    """
    discounts = pd.DataFrame([(r+1)**-i for i in t])
    discounts.index = t
    return discounts


import random
import pandas as pd
import numpy as np
import yfinance
import matplotlib.lines as mlines
import matplotlib.pyplot as plt
from scipy.stats import kurtosis
plt.rcParams['axes.facecolor'] = 'lightblue'


def create_ts(params={'data': np.array([]), 'timestamps': pd.Series([1])}):
    '''
    Input paramters:
    ------------------------------------------------
    params: dict (dictionary) containing the dates with column named "timestamps"
            and the data containing the data corresponding to the dates.

    >>> Example:
    >>> create_ts(params={'data': np.array([1, 2, 3]), 'timestamps': pd.Series(['2017-01-01', '2018-02-01', '2016-05-07'])})

    Returns: pd.Series with date as index and data.
    '''
    return pd.Series(params['data'], index=params['timestamps'].values)


def get_signal_strategy(stock_short, stock_long):
    '''
    Input: Takes 2 parameters : short time period moving average, long time period moving average.

    Returns: Digital Signal of the Strategy
    '''
    arr = stock_short - stock_long
    signal = np.sign(arr)
    signal = np.nan_to_num(signal)

    return signal


def split_func(X, test_size=0.2):
    '''
    Input: An array or list
    test_size = 0.2 (the size of test set)
                default value is 20%.
    '''
    train_size = 1 - test_size
    return X[:int(train_size*len(X))], X[int(train_size*len(X)):]


def get_strategy_aroon(up, down):
    '''
    Input: up , down signals of Aroon indicator
    ---------------------------------------------------------------
    Parameters:
              up: Array Like
              down: Array Like
    Returns: digital output of the signal .
    '''

    signal = [0, 0]
    count = 0

    for i in range(2, len(up)):
        if up[i-2] < down[i-2] and up[i] > down[i]:
            signal.append(1)
        elif down[i-2] < up[i-2] and down[i] > up[i]:
            signal.append(-1)
        elif up[i] > 50 and down[i] < 50:
            signal.append(1)
        elif up[i] < 50 and down[i] > 50:
            signal.append(-1)
        else:
            signal.append(0)

    return signal


def sigmoid(x):
    '''
    input: Accepts a list or array like input
    outputs: scaled value between 0 and 1.

    >>> Example:
    >>> sigmoid([1])
    >>> [0.73]
    '''
    return np.around(1/(1 + np.e**(-x)), 2)


def digital_signal(super_indicator):
    '''
    Input: Given a list or array like input converts it into a digital output

    Example:
    >>> digital_signal([1, 2, 3, 4, 5, 6, 7])
    >>> [1, 1, 1, 1, 1, 1, 1]
    '''
    signal = []

    for i in super_indicator:
        j = sigmoid(i)
        if j >= 0.5:
            signal.append(1)
        else:
            signal.append(-1)

    return signal


def signal_strategy_bollb(up, down, price):
    '''
    Input: Takes 3 parameters : up, down and price (list or array)
    The 3 bands of the bollinger band
    ---------------------------------------------------------------
    Parameters:
              up: Array Like
              down: Array Like
              price: Array Like

    Returns: Digital Signal of the Strategy
    '''

    signal = []

    for i in range(len(price)):
        if (up[i] - price[i]) > (price[i] - down[i]):
            signal.append(1)
        elif (up[i] - price[i]) - (price[i] - down[i]) < 1:
            signal.append(0)
        else:
            signal.append(-1)

    return signal


def financial_summary(df_rets, frequency='D'):
    '''
    Must supply a dataframe with date and daily retruns as columns
    Note - Don't supply daily returns as % . Keep the date column at the beginning.
    Example:
    date           returns
    2018-02-09     0.25
    2018-02-10     0.29

    frequency : Daily (D), Monthly(M) or Weekly(W), default - 'D'.
                Describes the frequency of data provided.

    Outputs: pd.DataFrame()
    '''
    if frequency == 'D':
        df_rets['c_ret'] = (1 + df_rets['returns']).cumprod() - 1
        days = (pd.to_datetime(df_rets.iloc[len(df_rets)-1, 0]) - pd.to_datetime(df_rets.iloc[0, 0])) // np.timedelta64(1, 'D')
        volatility = np.std(df_rets['returns']) * np.sqrt(252)
        returns = ((df_rets['c_ret'].values[-1])/(days)) * 252
        sharpe = returns/volatility
        sortino = returns / (np.std(df_rets[df_rets['returns'] < 0]['returns']) * 252)

        return pd.DataFrame(data=[df_rets.iloc[0, 0],
                           df_rets.iloc[len(df_rets)-1, 0],
                           days,
                           np.around(returns*100, 2),
                           np.around(volatility*100, 2),
                           np.around(sharpe, 2),
                           np.around(kurtosis(df['returns'], fisher=False), 2),
                           np.around(drawdown(df['returns'])['Drawdown'].min()*100, 2),
                           np.around(sortino, 2)],

                          columns=['Summary'],
                          index=['Start Date', 'End Date', 'Time Period (in Days)', 'Annual Return %',
                                 'Annual Volatility %', 'Sharpe Ratio', 'Kurtosis', 'Max Drawdown %', 'Sortino Ratio'])

    elif frequency == 'M':
        df_rets['c_ret'] = (1 + df_rets['returns']).cumprod() - 1
        months = (pd.to_datetime(df_rets.iloc[len(df_rets)-1, 0]) - pd.to_datetime(df_rets.iloc[0, 0])) // np.timedelta64(1, 'M')
        volatility = np.std(df_rets['returns']) * np.sqrt(12)
        returns = ((df_rets['c_ret'].values[-1])/(months)) * 12
        sharpe = returns/volatility
        sortino = returns / (np.std(df_rets[df_rets['returns'] < 0]['returns']) * 12)

        return pd.DataFrame(data=[df_rets.iloc[0, 0],
                           df_rets.iloc[len(df_rets)-1, 0],
                           months,
                           np.around(returns*100, 2),
                           np.around(volatility*100, 2),
                           np.around(sharpe, 2),
                           np.around(kurtosis(df['returns'], fisher=False), 2),
                           np.around(drawdown(df['returns'])['Drawdown'].min()*100, 2),
                           np.around(sortino, 2)],

                          columns=['Summary'],
                          index=['Start Date', 'End Date', 'Time Period (in Months)', 'Annual Return %',
                                 'Annual Volatility %', 'Sharpe Ratio', 'Kurtosis', 'Max Drawdown %', 'Sortino Ratio'])

    elif frequency == 'W':
        df_rets['c_ret'] = (1 + df_rets['returns']).cumprod() - 1
        weeks = (pd.to_datetime(df_rets.iloc[len(df_rets)-1, 0]) - pd.to_datetime(df_rets.iloc[0, 0])) // np.timedelta64(1, 'W')
        volatility = np.std(df_rets['returns']) * np.sqrt(52)
        returns = ((df_rets['c_ret'].values[-1])/(weeks)) * 52
        sharpe = returns/volatility
        sortino = returns / (np.std(df_rets[df_rets['returns'] < 0]['returns']) * 52)

        return pd.DataFrame(data=[df_rets.iloc[0, 0],
                           df_rets.iloc[len(df_rets)-1, 0],
                           weeks,
                           np.around(returns*100, 2),
                           np.around(volatility*100, 2),
                           np.around(sharpe, 2),
                           np.around(kurtosis(df['returns'], fisher=False), 2),
                           np.around(drawdown(df['returns'])['Drawdown'].min()*100, 2),
                           np.around(sortino, 2)],

                          columns=['Summary'],
                          index=['Start Date', 'End Date', 'Time Period (in Weeks)', 'Annual Return %',
                                 'Annual Volatility %', 'Sharpe Ratio', 'Kurtosis', 'Max Drawdown %', 'Sortino Ratio'])


def drawdown(return_series: pd.Series):
    """Takes a time series of asset returns.
       returns a DataFrame with columns for
       the wealth index,
       the previous peaks, and
       the percentage drawdown
    """
    wealth_index = 1000*(1+return_series).cumprod()
    previous_peaks = wealth_index.cummax()
    drawdowns = (wealth_index - previous_peaks)/previous_peaks
    return pd.DataFrame({"Wealth": wealth_index,
                         "Previous Peak": previous_peaks,
                         "Drawdown": drawdowns})

def analyze_strategy(actions, df, frequency='D', plot=False):
    '''
    actions: a list , np.array or series of actions .
    plot: default-False, boolean
    df: prices_with_dates - a list of prices along with their respective dates (pd.DataFrame)
                            name of the price column should be "prices"
                            Example :
                            Date           prices
                            "2017-01-09"   22.87

    frequency : Daily (D), Monthly(M) or Weekly(W), default - 'D'.
                Describes the frequency of data provided.

    Example:
    >>> analyze_strategy([1, 1, 1, ..-1], df, 'D', True)
    actions : +1 , 0 or -1 . (Buy, Hold or Sell)

    outputs summary for the actions taken .
    '''

    df['actions'] = actions
    df['returns'] = df['prices'].pct_change() * df['actions'].shift(1)
    df['c_ret'] = (1 + df['returns']).cumprod() - 1
    df.fillna(0, inplace=True)

    if frequency == 'D':
        days = (pd.to_datetime(df.iloc[len(df)-1, 0]) - pd.to_datetime(df.iloc[0, 0]))//np.timedelta64(1, 'D')
        volatility = np.std(df['returns']) * np.sqrt(252)
        returns = ((df['c_ret'].values[-1])/(days)) * 252
        sharpe = returns/volatility
    elif frequency == 'M':
        months = (pd.to_datetime(df.iloc[len(df)-1, 0]) - pd.to_datetime(df.iloc[0, 0]))//np.timedelta64(1, 'M')
        volatility = np.std(df['returns']) * np.sqrt(12)
        returns = ((df['c_ret'].values[-1])/(months)) * 12
        sharpe = returns/volatility
    elif frequency == 'W':
        weeks = (pd.to_datetime(df.iloc[len(df)-1, 0]) - pd.to_datetime(df.iloc[0, 0]))//np.timedelta64(1, 'W')
        volatility = np.std(df['returns']) * np.sqrt(52)
        returns = ((df['c_ret'].values[-1])/(weeks)) * 52
        sharpe = returns/volatility

    if plot:
        dq = df.set_index(df.columns[0])
        plt.title("Returns over the years")
        dq['c_ret'].plot(figsize=(12, 6))
        plt.xlabel("Days")
        plt.ylabel("Returns (Compounded)")


    return pd.DataFrame(data=[df.iloc[0, 0],
                           df.iloc[len(df)-1, 0],
                           days,
                           np.around(returns*100, 2),
                           np.around(volatility*100, 2),
                           np.around(sharpe, 2),
                           np.around(kurtosis(df['returns'], fisher=False), 2),
                           np.around(drawdown(df['returns'])['Drawdown'].min()*100, 2)],

                          columns=['Summary'],
                          index=['Start Date', 'End Date', f'Time Period (in ({frequency}))', 'Annual Return %',
                                 'Annual Volatility %', 'Sharpe Ratio', 'Kurtosis', 'Max Drawdown %'])


def get_benchmark_result(ticker, start, end, price_col='Adj Close', frequency='D', plot=False):
    '''
    start: start date
    end: end date
    ticker: symbol (of Benchmark index eg: ^NSEI for NIFTY)
    plot: default-False, boolean

    frequency : Daily (D), Monthly(M) or Weekly(W), default - 'D'.
                Describes the frequency of data provided.

    price : Open, High , Low or Close . Default = 'Close' .

    returns summary for the benchmark
    '''
    data = yfinance.download(tickers=ticker, start=start, end=end, interval='1d', progress=False)
    mapper = {"D": "Days", "M": "Monthly", "W": "Weeks"}

    if frequency == 'D':
        df = data.loc[:, price_col].reset_index().rename(columns={price_col: "prices"})
        df['returns'] = df['prices'].pct_change()
        df['c_ret'] = (1 + df['returns']).cumprod() - 1
        df.fillna(0, inplace=True)

        days = (pd.to_datetime(df.iloc[len(df)-1, 0]) - pd.to_datetime(df.iloc[0, 0]))//np.timedelta64(1, 'D')
        volatility = np.std(df['returns']) * np.sqrt(252)
        returns = ((df['c_ret'].values[-1])/(days)) * 252
        sharpe = returns/volatility

    elif frequency == 'M':
        data = data.resample('M').last()
        df = data.loc[:, price_col].reset_index().rename(columns={price_col: "prices"})
        df['returns'] = df['prices'].pct_change()
        df['c_ret'] = (1 + df['returns']).cumprod() - 1
        df.fillna(0, inplace=True)

        days = (pd.to_datetime(df.iloc[len(df)-1, 0]) - pd.to_datetime(df.iloc[0, 0]))//np.timedelta64(1, 'M')
        volatility = np.std(df['returns']) * np.sqrt(12)
        returns = ((df['c_ret'].values[-1])/(days)) * 12
        sharpe = returns/volatility

    elif frequency == 'W':
        data = data.resample('W').last()
        df = data.loc[:, price_col].reset_index().rename(columns={price_col: "prices"})
        df['returns'] = df['prices'].pct_change()
        df['c_ret'] = (1 + df['returns']).cumprod() - 1
        df.fillna(0, inplace=True)

        days = (pd.to_datetime(df.iloc[len(df)-1, 0]) - pd.to_datetime(df.iloc[0, 0]))//np.timedelta64(1, 'W')
        volatility = np.std(df['returns']) * np.sqrt(52)
        returns = ((df['c_ret'].values[-1])/(days)) * 52
        sharpe = returns/volatility

    if plot:
        dq = df.set_index("Date")
        plt.title("Returns over the years")
        dq['c_ret'].plot(figsize=(12, 6))
        plt.xlabel("Days")
        plt.ylabel("Returns (Compounded)")


    return pd.DataFrame(data=[str(df.iloc[0, 0])[:10],
                           str(df.iloc[len(df)-1, 0])[:10],
                           days,
                           np.around(returns*100, 2),
                           np.around(volatility*100, 2),
                           np.around(sharpe, 2),
                           np.around(kurtosis(df['returns'], fisher=False), 2),
                           np.around(drawdown(df['returns'])['Drawdown'].min()*100, 2)],

                          columns=['Summary'],
                          index=['Start Date', 'End Date', f'Time Period (in ({frequency}))', 'Annual Return %',
                                 'Annual Volatility %', 'Sharpe Ratio', 'Kurtosis', 'Max Drawdown %'])


def MDD(rets):
    '''
    Input: Takes a return series and returns the maximum drawdown.
    '''
    final = rets
    maxdrwdn = (1+final).cumprod().diff().min()

    return maxdrwdn


def detailed_summary(excelSheet, sheets, period=12, prev_yr='2017'):
    '''
    Provides an year-wise summary of portfolio, baseline and the individual stocks.

    excelSheet: Name of the ExcelSheet with the details.
    sheets: A list with sheet Names - ["Portfolio", "Baseline", "Prices"]
    period: Data details, whether monthly, daily or quarterly data. Default is 12 i.e. monthly, change it to 4 if quarterly or 252 if daily.
    prev_yr: last year preceeding the first date. For Ex: if first date is "2018-03-01", prev_yr = 2017

    Note: "Provide the entire path to the excelsheet if it is inside a folder."
          All sheets must have a column named "date" for dates.
          "invested" - Investments in portfolio (in Portfolio sheet).
          "Adj Close" - In Baseline Sheet.(In Baseline sheet)
    '''

    def individual_etfs(df):
        Y = list(df.columns[1:len(df.columns)-1])
        dfx = pd.DataFrame()

        # for all etfs in the portfolio
        for etf in Y:
            tf = pd.DataFrame()
            prev_date = prev_yr
            for date in df['year'].unique():
                y = df[(df_port.date>=prev_date+'-12-31') & (df_port.date<=date+'-12-31')].loc[:, [etf]]
                y.reset_index(drop=True, inplace=True)

                # Date update
                prev_date = date

                # Log Returns
                lgrets = np.diff(np.log(y[etf]))
                lgrets = np.insert(lgrets, 0, np.nan)
                y['log_returns'] = lgrets
                y['log_returns'].fillna(0, inplace=True)

                # Volatility
                vol = np.std(y.loc[1:, 'log_returns'], ddof=1) * np.sqrt(period)

                # Annual Return
                ret = (y.loc[len(y)-1, etf] / y.loc[0, etf]) - 1

                # Max Drawdown
                inv = y[etf]
                z = pd.Series(index=range(len(inv)))
                z.iloc[0] = inv.iloc[0]

                for i in range(1, len(inv)):
                    z.iloc[i] = max(inv[i], z[i-1])

                maxdrwdn = (inv - z).min()/z[0]

                # Sharpe ratio
                sharpe = ret / vol

                new_df = pd.DataFrame(data=[ret, vol, sharpe, maxdrwdn], columns=[date], index=['Return', 'Volatility', 'Sharpe Ratio', 'Max Drawdown'])
                tf = pd.concat([tf, new_df], axis=1)

            tf.columns.name = etf
            tf['Security Name'] = etf
            dfx = pd.concat([dfx, tf])

        return dfx


    # Read the entire excel sheet
    result = pd.ExcelFile(excelSheet)
    df = pd.DataFrame()
    prev_date = prev_yr

    for ticker in sheets:
        df_port = pd.read_excel(result, ticker)
        df_port['year'] = df_port['date'].apply(lambda x: x[:4])
        tf = pd.DataFrame()

        if ticker == 'Portfolio':
            for date in df_port['year'].unique():
                y = df_port[(df_port.date>=prev_date+'-12-31') & (df_port.date<=date+'-12-31')]
                y.reset_index(drop=True, inplace=True)

                # Date update
                prev_date = date

                # Log Returns
                lgrets = np.diff(np.log(y['invested']))
                lgrets = np.insert(lgrets, 0, np.nan)
                y['log_returns'] = lgrets
                y['log_returns'].fillna(0, inplace=True)

                # Volatility
                vol = np.std(y.loc[1:, 'log_returns'], ddof=1) * np.sqrt(period)

                # Annual Return
                ret = (y.loc[len(y)-1, 'invested'] / y.loc[0, 'invested']) - 1

                # Max Drawdown
                inv = y['invested']
                z = pd.Series(index=range(len(inv)))
                z.iloc[0] = inv.iloc[0]

                for i in range(1, len(inv)):
                    z.iloc[i] = max(inv[i], z[i-1])

                maxdrwdn = (inv - z).min()/z[0]

                # Sharpe ratio
                sharpe = ret / vol

                new_df = pd.DataFrame(data=[ret, vol, sharpe, maxdrwdn], columns=[date], index=['Return', 'Volatility', 'Sharpe Ratio', 'Max Drawdown'])
                tf = pd.concat([tf, new_df], axis=1)

            tf.columns.name = ticker
            tf['Security Name'] = ticker
            df = pd.concat([df, tf])

        elif ticker == 'Baseline':
            prev_date = prev_yr

            for date in df_port['year'].unique():
                y = df_port[(df_port.date>=prev_date+'-12-31') & (df_port.date<=date+'-12-31')]
                y.reset_index(drop=True, inplace=True)

                # Date update
                prev_date = date

                # Log Returns
                lgrets = np.diff(np.log(y['Adj Close']))
                lgrets = np.insert(lgrets, 0, np.nan)
                y['log_returns'] = lgrets
                y['log_returns'].fillna(0, inplace=True)

                # Volatility
                vol = np.std(y.loc[1:, 'log_returns'], ddof=1) * np.sqrt(period)

                # Annual Return
                ret = (y.loc[len(y)-1, 'Adj Close'] / y.loc[0, 'Adj Close']) - 1

                # Max Drawdown
                inv = y['Adj Close']
                z = pd.Series(index=range(len(inv)))
                z.iloc[0] = inv.iloc[0]

                for i in range(1, len(inv)):
                    z.iloc[i] = max(inv[i], z[i-1])

                maxdrwdn = (inv - z).min()/z[0]

                # Sharpe ratio
                sharpe = ret / vol

                new_df = pd.DataFrame(data=[ret, vol, sharpe, maxdrwdn], columns=[date], index=['Return', 'Volatility', 'Sharpe Ratio', 'Max Drawdown'])
                tf = pd.concat([tf, new_df], axis=1)

            tf.columns.name = ticker
            tf['Security Name'] = ticker
            df = pd.concat([df, tf])

        else:
            tf = individual_etfs(df_port)
            df = pd.concat([df, tf])

    df.columns.name = 'Metric'
    return df



from .plots import *
from .technical_indicators import *
from .optimizers import *
from .new_optimizers import *
from .utils import *
from .statistical_tests import *
from .finstats import *
from .chartpack import *



from setuptools import setup, find_packages

try:
    REQUIRES = list()
    f = open("requirements.txt", "rb")
    for line in f.read().decode("utf-8").split("\n"):
        line = line.strip()
        if "#" in line:
            line = line[: line.find("#")].strip()
        if line:
            REQUIRES.append(line)
except FileNotFoundError:
    print("'requirements.txt' not found!")
    REQUIRES = list()

url_path = ''

setup(name='financiallib',
      version = '1.4.8',
      license = f"Financial Services{url_path}",
      author = "Neelakash Chatterjee",
      packages = find_packages(),
      install_requires = REQUIRES,
      description = "Financial-Lib: A Library for Financial Data Analysis and Pre-Processing.",
      )



from .technical_indicators import *
from .plots import *
from .utils import *
import random

def sma_optimizer(price, mins, maxs, minl, maxl):
    """
    Pass a list of prices (or even an array)
    mins, maxs, minl, maxl - are the minimum and maximum ranges for both short and long time periods for optimization.

    returns : a tuple of t1, t2 and the optimal signal (t1 and t2 are optimal time periods)
    """
    try:
        opt = -np.inf
        for i in range(mins, maxs):
            for j in range(minl, maxl):
                if i < j:
                    stock_short = simple_moving_average(price, {'periods': i})
                    stock_long = simple_moving_average(price, {'periods': j})

                    signal = get_signal_strategy(stock_short, stock_long)

                    pl, _, _ = agg_pl_max_dd(signal, price)

                    avg_pl = pl

                    if avg_pl > opt:
                        opt_signal = signal
                        opt = avg_pl
                        t1 = i
                        t2 = j

        return t1, t2, opt_signal
    except Exception as e:
        print(e)


def ema_optimizer(price, mins, maxs, minl, maxl):
    """
    Pass a list of prices (or even an array)
    mins, maxs, minl, maxl - are the minimum and maximum ranges for both short and long time periods for optimization.
    returns : a tuple of t1, t2 and the optimal signal (t1 and t2 are optimal time periods)
    """
    try:
        opt = -np.inf
        for i in range(mins, maxs):
            for j in range(minl, maxl):
                if i < j:
                    stock_short = exponential_moving_average(price, {'periods': i})
                    stock_long = exponential_moving_average(price, {'periods': j})

                    signal = get_signal_strategy(stock_short, stock_long)

                    pl, _, _ = agg_pl_max_dd(signal, price)

                    avg_pl = pl

                    if avg_pl > opt:
                        opt_signal = signal
                        opt = avg_pl
                        t1 = i
                        t2 = j

        return t1, t2, opt_signal
    except Exception as e:
        print(e)


def aroon_optimizer(price, minr, maxr):
    """
    Pass a list of prices (or even an array)
    minr, maxr - are the minimum and maximum ranges for  time periods for optimization.
    returns : a tuple of t1, and the optimal signal (t1 is optimal time period)
    """
    try:
        opt = -np.inf
        for i in range(minr, maxr):
            up, down = aroon_indicator(price, {'lb': i})

            signal = get_strategy_aroon(up, down)

            pl, _, _ = agg_pl_max_dd(signal, price)

            avg_pl = pl

            if avg_pl > opt:
                opt_signal = signal
                opt = avg_pl
                t1 = i

        return t1, opt_signal
    except Exception as e:
        print(e)


def bollinger_optimizer(prices, minr, maxr):
    """
    Pass a list of prices (or even an array)
    minr, maxr - are the minimum and maximum ranges for  time periods for optimization.
    returns : a tuple of t1, and the optimal signal (t1 is optimal time period)
    """
    try:
        opt = -np.inf
        for i in range(minr, maxr):
            up, down, price = get_bollinger_bands(prices, {'rate': i})

            signal = signal_strategy_bollb(up, down, price)

            pl, _, _ = agg_pl_max_dd(signal, prices)

            avg_pl = pl

            if avg_pl > opt:
                opt_signal = signal
                opt = avg_pl
                t1 = i

        return t1, opt_signal
    except Exception as e:
        print(e)


def gods_signal(price):
    """
    pass a list or an array of prices.

    outputs: god signal

    God Signal is the desired or the ultimate signal , the position one would take
    if they come to know about the price tomorrow.
    """
    try:
        price = pd.Series(price)
        god_signal = np.sign(price.diff())
        return np.nan_to_num(god_signal)

    except Exception as e:
        print(e)



import pandas as pd
import matplotlib.pyplot as plt
import numpy as np
import warnings
warnings.filterwarnings("ignore")
from pykalman import KalmanFilter

def simple_moving_average(arr, params={'periods': 2}):
    """
    Parameters
    -------------------------------------------------
    arr: np.array, list or pd.Series
    params: dict (with 'periods' as key and an integer as value)
            value is the window size , default value is 2.

    returns: np.array (SMA signal)
    """
    try:
        new_data = np.array(arr)
        data = pd.DataFrame(new_data, columns=['value'])
        data['Rolling Mean'] = data['value'].rolling(window=params['periods']).mean()
        mov_avg = np.array(data['Rolling Mean'])

        return mov_avg

    except Exception as e:
        print(e)


def relative_strength_index(arr, params={'periods': 14, 'kind': simple_moving_average}):
    """
    Parameters
    -------------------------------------------------
    arr: np.array, list or pd.Series
    params: dict (with 'periods' as key and an integer as value
            value is the window size , default value is 14. The second key 
            here is the 'kind', which is by default 'SMA')

    returns: rsi signal 
    """
    try:
        new_data = np.array(arr)
        data = pd.DataFrame(new_data, columns=['values'])
        data['diff'] = data['values'].diff()
        dUp, dDown = data['diff'].copy(), data['diff'].copy()

        dUp[dUp < 0] = 0
        dDown[dDown > 0] = 0

        dDown = dDown.abs()

        mUp = params['kind'](dUp, {'periods': params['periods']})
        mDown = params['kind'](dDown, {'periods': params['periods']})

        rs = mUp / mDown
        rsi = 100 - (100/(1 + rs))

        return rsi

    except Exception as e:
        print(e)


def exponential_moving_average(arr, params = {'periods': 9}):
    """
    Parameters
    -------------------------------------------------
    arr: np.array, list or pd.Series
    params: dict (with 'periods' as key and an integer as value)
            value is the window size , default value is 9.

    returns: np.array (EMA signal)
    """
    try:
        arr1 = np.array(arr)
        arr2 = pd.DataFrame(arr1, columns = ['value'])
        arr2['ewm'] = arr2['value'].ewm(span = params['periods'], adjust=False).mean()
        exp_avg = np.array(arr2['ewm'])

        return exp_avg

    except Exception as e:
        print(e)


def ROC(arr, params={'periods': 2}):
    """
    Parameters
    -------------------------------------------------
    arr: np.array, list or pd.Series
    params: dict (with 'periods' as key and an integer as value)
            value is the window size , default value is 2.

    returns: np.array (ROC signal)
    """
    try:
        arr = np.array(arr)
        data = pd.DataFrame(arr, columns=['values'])
        N = data['values'].diff(params['periods'])
        D = data['values'].shift(params['periods'])
        ROC = pd.Series(N/D, name='ROC')

        return np.array(ROC)
    except Exception as e:
        print(e)


def get_bollinger_bands(arr, params={'rate': 20}):
    """
    Parameters
    -------------------------------------------------
    arr: np.array, list or pd.Series
    params: dict (with 'rate' as key and an integer as value)
            value is the window size , default value is 20.

    returns: 3 bands of bollinger .
    """
    try:
        arr = np.array(arr)
        data = pd.DataFrame(arr, columns=['values'])
        sma = simple_moving_average(arr, {'periods': params['rate']})
        data['std'] = data['values'].rolling(params['rate']).std()
        bollinger_up = sma + (data['std'] * 2)
        bollinger_down = sma - (data['std'] * 2)

        return np.array(bollinger_up), np.array(bollinger_down), sma
    except Exception as e:
        print(e)


def get_macd(arr, params={'fast': 12, 'slow': 26, 'signal-span': 9, 'adjust': False}):
    """
    Parameters
    -------------------------------------------------
    arr: np.array, list or pd.Series
    params: dict [keys and defaults - 'fast': 12, 'slow': 26, 'signal-span': 9, 'adjust': False]

    returns: np.array (MACD, Signal) a tuple of two numpy arrays
    """
    try:
        arr = np.array(arr)
        data = pd.DataFrame(arr, columns=['values'])
        exp1 = data['values'].ewm(span=params['fast'], adjust=params['adjust']).mean()
        exp2 = data['values'].ewm(span=params['slow'], adjust=params['adjust']).mean()
        data['macd'] = exp1 - exp2
        data['signal'] = data['macd'].ewm(span=params['signal-span'], adjust=params['adjust']).mean()

        return np.array(data['macd']), np.array(data['signal'])
    except Exception as e:
        print(e)


def on_balance_volume(arr, params = {'volume': [4, 7, 9, 10]}):
    """
    Parameters
    -------------------------------------------------
    arr: np.array, list or pd.Series
    params: dict (with 'volume' as key and a list as value)
            by default a dummy list is passed make sure to pass a 
            valid list or array.

    returns: np.array (OBV signal)
    """
    try:
        arr1 = np.array(arr)
        arr3 = np.array(params['volume'])
        arr2 = pd.DataFrame(arr1, columns=['values'])
        arr4 = pd.DataFrame(arr3, columns=['volume'])
        data = pd.concat([arr2, arr4], axis=1)

        data['on_bal_val'] = (np.sign(data['values'].diff()) * data['volume']).fillna(0).cumsum()
        obv_avg = np.array(data['on_bal_val'])
        return obv_avg

    except Exception as e:
        print(e)


def aroon_indicator(arr, params = {'lb' : 25}):
    """
    Parameters
    -------------------------------------------------
    arr: np.array, list or pd.Series
    params: dict (with 'lb' as key and an integer as value)
            default value is 25.

    returns: np.array (Aroon signal)
    """
    try:
        data = np.array(arr)
        new_data = pd.DataFrame(data, columns=['values'])

        df = new_data.copy()
        df['up'] = 100 * df['values'].rolling(params['lb'] + 1).apply(lambda x: x.argmax()) / params['lb']
        df['down'] = 100 * df['values'].rolling(params['lb'] + 1).apply(lambda x: x.argmin()) / params['lb']

        return np.array(df['up']), np.array(df['down'])

    except Exception as e:
        print(e)


def stochastic(arr, high, low, params={'%K-type': 'slow'}):
    """
    Parameters
    -------------------------------------------------
    arr: np.array, list or pd.Series
    high: np.array, list or pd.Series
    low: np.array, list or pd.Series
    params: dict (with '%K-type' as key and a string as value)
            value is 'slow' by default , may use 'fast'.

    returns: np.array (tuple) , Stochastic Oscillator
    """
    if params['%K-type'] == 'slow':
        n = 3
    elif params['%K-type'] == 'fast':
        n = 1
    else:
        print("Enter only fast or slow .")

    try:
        arr, high, low = np.array(arr), np.array(high), np.array(low)
        df = {'values': arr, 'High': high, 'Low': low}
        data = pd.DataFrame(df, columns=['values', 'High', 'Low'])
        data['14-high'] = data['High'].rolling(14).max()
        data['14-low'] = data['Low'].rolling(14).min()
        data['%K'] = (data['values'] - data['14-low'])*100/(data['14-high'] - data['14-low'])
        data['%D'] = data['%K'].rolling(n).mean()

        return np.array(data['%K']), np.array(data['%D'])
    except Exception as e:
        print(e)


def kalmanFilter(arr):
    """
    Parameters
    -------------------------------------------------
    arr: np.array, list or pd.Series
    
    returns: np.array (tuple) (kf_mean and Covariance)
    """
    
    kf = KalmanFilter(transition_matrices = [1], observation_matrices = [1], initial_state_mean = 0,
                initial_state_covariance = 1, observation_covariance = 1, transition_covariance =.01)

    try:
        arr = np.array(arr)
        state_means_kf, covar = kf.filter(arr)

        return np.array(state_means_kf), np.array(covar)

    except Exception as e:
        print(e)


def agg_pl_max_dd(signal, price, plot=False, god=False):
    """
    Parameters
    ----------------------------------------------------
    signal: a list or array of position values 
    price: a list or array of price values
    plot: True or False, default:False
    god: False (default), True only if it's a god signal

    returns: tuple of (cumulative p/L, minimum of cum_p/L, a dataframe)
    """
    try:
        if god:
            signal, price = np.array(signal), np.array(price)
            data = pd.DataFrame({'Signal': signal, 'Price': price})
            data['Price_Change'] = data['Price'].diff()
            data['Cumulative_P/L'] = (data['Price_Change'] * data['Signal']).cumsum()

        else:
            signal, price = np.array(signal), np.array(price)
            data = pd.DataFrame({'Signal': signal, 'Price': price})
            data['Price_Change'] = data['Price'].diff()
            data['Cumulative_P/L'] = (data['Price_Change'] * data['Signal'].shift(1)).cumsum()

        if plot:
            plt.figure(figsize=(15, 8))
            plt.title("Maximum Drawdown and Aggregate P/L", fontsize=17)
            plt.plot(data['Price'], color='r', linewidth=2)
            plt.plot(data['Price_Change'], color='b', linewidth=2)
            plt.plot(data['Cumulative_P/L'], color='g', linewidth=2)
            plt.xlabel('Days', fontsize=15)
            plt.ylabel('Value', fontsize=15)
            plt.legend(('Price', 'Price_Change', 'Cumulative_P/L'))

        return data['Cumulative_P/L'][len(data)-1], data['Cumulative_P/L'].min(), data
    except Exception as e:
        print(e)
